#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This is a simple echo bot using decorators and webhook with BaseHTTPServer
# It echoes any incoming text messages and does not use the polling method.

try:
    # Python 2
    from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
except ImportError:
    # Python 3
    from http.server import BaseHTTPRequestHandler, HTTPServer

import logging
import ssl

import telebot

API_TOKEN = "<api_token>"

WEBHOOK_HOST = "<ip/host where the bot is running>"
WEBHOOK_PORT = 8443  # 443, 80, 88 or 8443 (port need to be 'open')
WEBHOOK_LISTEN = "0.0.0.0"  # In some VPS you may need to put here the IP addr

WEBHOOK_SSL_CERT = "./webhook_cert.pem"  # Path to the ssl certificate
WEBHOOK_SSL_PRIV = "./webhook_pkey.pem"  # Path to the ssl private key

# Quick'n'dirty SSL certificate generation:
#
# openssl genrsa -out webhook_pkey.pem 2048
# openssl req -new -x509 -days 3650 -key webhook_pkey.pem -out webhook_cert.pem
#
# When asked for "Common Name (e.g. server FQDN or YOUR name)" you should reply
# with the same value in you put in WEBHOOK_HOST

WEBHOOK_URL_BASE = "https://%s:%s" % (WEBHOOK_HOST, WEBHOOK_PORT)
WEBHOOK_URL_PATH = "/%s/" % (API_TOKEN)

logger = telebot.logger
telebot.logger.setLevel(logging.INFO)

bot = telebot.TeleBot(API_TOKEN)


# WebhookHandler, process webhook calls
class WebhookHandler(BaseHTTPRequestHandler):
    server_version = "WebhookHandler/1.0"

    def do_HEAD(self):
        self.send_response(200)
        self.end_headers()

    def do_GET(self):
        self.send_response(200)
        self.end_headers()

    def do_POST(self):
        if (
            self.path == WEBHOOK_URL_PATH
            and "content-type" in self.headers
            and "content-length" in self.headers
            and self.headers["content-type"] == "application/json"
        ):
            json_string = self.rfile.read(int(self.headers["content-length"]))

            self.send_response(200)
            self.end_headers()

            update = telebot.types.Update.de_json(json_string)
            bot.process_new_messages([update.message])
        else:
            self.send_error(403)
            self.end_headers()


# Handle '/start' and '/help'
@bot.message_handler(commands=["help", "start"])
def send_welcome(message):
    bot.reply_to(
        message,
        ("Hi there, I am EchoBot.\n" "I am here to echo your kind words back to you."),
    )


# Handle all other messages
@bot.message_handler(func=lambda message: True, content_types=["text"])
def echo_message(message):
    bot.reply_to(message, message.text)


# Remove webhook, it fails sometimes the set if there is a previous webhook
# bot.remove_webhook()

# Set webhook
# Beacuse telegram bot api server will check webhook server is alive.
# Here we need set webhook after server started manually.
# bot.set_webhook(url=WEBHOOK_URL_BASE + WEBHOOK_URL_PATH,
#                certificate=open(WEBHOOK_SSL_CERT, 'r'))

# Start server
httpd = HTTPServer((WEBHOOK_LISTEN, WEBHOOK_PORT), WebhookHandler)

httpd.socket = ssl.wrap_socket(
    httpd.socket, certfile=WEBHOOK_SSL_CERT, keyfile=WEBHOOK_SSL_PRIV, server_side=True
)

httpd.serve_forever()



import telebot
from telebot.async_telebot import AsyncTeleBot

bot = AsyncTeleBot("TOKEN")


# AdvancedCustomFilter is for list, string filter values
class MainFilter(telebot.asyncio_filters.AdvancedCustomFilter):
    key = "text"

    @staticmethod
    async def check(message, text):
        return message.text in text


# SimpleCustomFilter is for boolean values, such as is_admin=True
class IsAdmin(telebot.asyncio_filters.SimpleCustomFilter):
    key = "is_admin"

    @staticmethod
    async def check(message: telebot.types.Message):
        result = await bot.get_chat_member(message.chat.id, message.from_user.id)
        return result.status in ["administrator", "creator"]


@bot.message_handler(is_admin=True, commands=["admin"])  # Check if user is admin
async def admin_rep(message):
    await bot.send_message(message.chat.id, "Hi admin")


@bot.message_handler(is_admin=False, commands=["admin"])  # If user is not admin
async def not_admin(message):
    await bot.send_message(message.chat.id, "You are not admin")


@bot.message_handler(text=["hi"])  # Response to hi message
async def welcome_hi(message):
    await bot.send_message(message.chat.id, "You said hi")


@bot.message_handler(text=["bye"])  # Response to bye message
async def bye_user(message):
    await bot.send_message(message.chat.id, "You said bye")


# Do not forget to register filters
bot.add_custom_filter(MainFilter())
bot.add_custom_filter(IsAdmin())

import asyncio

asyncio.run(bot.polling())



import asyncio

import config
from aiohttp import web
from handlers import register_handlers

from telebot import types, util
from telebot.async_telebot import AsyncTeleBot

main_bot = AsyncTeleBot(config.MAIN_BOT_TOKEN)
app = web.Application()
tokens = {config.MAIN_BOT_TOKEN: True}


async def webhook(request):
    token = request.match_info.get("token")
    if not tokens.get(token):
        return web.Response(status=404)

    if request.headers.get("content-type") != "application/json":
        return web.Response(status=403)

    json_string = await request.json()
    update = types.Update.de_json(json_string)
    if token == main_bot.token:
        await main_bot.process_new_updates([update])
        return web.Response()

    from_update_bot = AsyncTeleBot(token)
    register_handlers(from_update_bot)
    await from_update_bot.process_new_updates([update])
    return web.Response()


app.router.add_post("/" + config.WEBHOOK_PATH + "/{token}", webhook)


@main_bot.message_handler(commands=["add_bot"])
async def add_bot(message: types.Message):
    token = util.extract_arguments(message.text)
    tokens[token] = True

    new_bot = AsyncTeleBot(token)
    await new_bot.delete_webhook()
    await new_bot.set_webhook(f"{config.WEBHOOK_HOST}/{config.WEBHOOK_PATH}/{token}")

    await new_bot.send_message(message.chat.id, "Webhook was set.")


async def main():
    await main_bot.delete_webhook()
    await main_bot.set_webhook(
        f"{config.WEBHOOK_HOST}/{config.WEBHOOK_PATH}/{config.MAIN_BOT_TOKEN}"
    )
    web.run_app(app, host=config.WEBAPP_HOST, port=config.WEBAPP_PORT)


if __name__ == "__main__":
    asyncio.run(main())



import logging

import telebot
from telebot.async_telebot import AsyncTeleBot, ExceptionHandler

logger = telebot.logger
telebot.logger.setLevel(logging.DEBUG)  # Outputs debug messages to console.


class MyExceptionHandler(ExceptionHandler):
    async def handle(self, exception):
        logger.error(exception)


bot = AsyncTeleBot("TOKEN", exception_handler=MyExceptionHandler())


@bot.message_handler(commands=["photo"])
async def photo_send(message: telebot.types.Message):
    await bot.send_message(
        message.chat.id, "Hi, this is an example of exception handlers."
    )
    raise Exception("test")  # Exception goes to ExceptionHandler if it is set


import asyncio

asyncio.run(bot.polling())



from telebot import formatting
from telebot.async_telebot import AsyncTeleBot

bot = AsyncTeleBot("token")


@bot.message_handler(commands=["start"])
async def start_message(message):
    await bot.send_message(
        message.chat.id,
        # function which connects all strings
        formatting.format_text(
            formatting.mbold(message.from_user.first_name),
            formatting.mitalic(message.from_user.first_name),
            formatting.munderline(message.from_user.first_name),
            formatting.mstrikethrough(message.from_user.first_name),
            formatting.mcode(message.from_user.first_name),
            separator=" ",  # separator separates all strings
        ),
        parse_mode="MarkdownV2",
    )

    # just a bold text using markdownv2
    await bot.send_message(
        message.chat.id,
        formatting.mbold(message.from_user.first_name),
        parse_mode="MarkdownV2",
    )

    # html
    await bot.send_message(
        message.chat.id,
        formatting.format_text(
            formatting.hbold(message.from_user.first_name),
            formatting.hitalic(message.from_user.first_name),
            formatting.hunderline(message.from_user.first_name),
            formatting.hstrikethrough(message.from_user.first_name),
            formatting.hcode(message.from_user.first_name),
            # hide_link is only for html
            formatting.hide_link("https://telegra.ph/file/c158e3a6e2a26a160b253.jpg"),
            separator=" ",
        ),
        parse_mode="HTML",
    )

    # just a bold text in html
    await bot.send_message(
        message.chat.id,
        formatting.hbold(message.from_user.first_name),
        parse_mode="HTML",
    )


import asyncio

asyncio.run(bot.polling())



#!/usr/bin/python

# This is a simple echo bot using the decorator mechanism.
# It echoes any incoming text messages.

from telebot.async_telebot import AsyncTeleBot

bot = AsyncTeleBot("TOKEN")


# Handle '/start' and '/help'
@bot.message_handler(commands=["help", "start"])
async def send_welcome(message):
    await bot.reply_to(
        message,
        """\
Hi there, I am EchoBot.
I am here to echo your kind words back to you. Just say anything nice and I'll say the exact same thing to you!\
""",
    )


# Handle all other messages with content_type 'text' (content_types defaults to ['text'])
@bot.message_handler(func=lambda message: True)
async def echo_message(message):
    await bot.reply_to(message, message.text)


import asyncio

# only for versions 4.7.0+
asyncio.run(bot.polling(restart_on_change=True))



"""
In this example you will learn how to adapt your bot to different languages
Using built-in class I18N.
You need to install babel package 'https://pypi.org/project/Babel/'
Babel provides a command-line interface for working with message catalogs
After installing babel package you have a script called 'pybabel'
Too see all the commands open terminal and type 'pybabel --help'
Full description for pybabel commands can be found here: 'https://babel.pocoo.org/en/latest/cmdline.html'
Create a directory 'locales' where our translations will be stored
First we need to extract texts:
    pybabel extract -o locales/{domain_name}.pot --input-dirs .
{domain_name}.pot - is the file where all translations are saved
The name of this file should be the same as domain which you pass to I18N class
In this example domain_name will be 'messages'
For gettext (singular texts) we use '_' alias and it works perfect
You may also you some alias for ngettext (plural texts) but you can face with a problem that
your plural texts are not being extracted
That is because by default 'pybabel extract' recognizes the following keywords:
 _, gettext, ngettext, ugettext, ungettext, dgettext, dngettext, N_
To add your own keyword you can use '-k' flag
In this example for 'ngettext' i will assign double underscore alias '__'
Full command with pluralization support will look so:
    pybabel extract -o locales/{domain_name}.pot -k __:1,2 --input-dirs .
Then create directories with translations (get list of all locales: 'pybabel --list-locales'):
    pybabel init -i locales/{domain_name}.pot -d locales -l en
    pybabel init -i locales/{domain_name}.pot -d locales -l ru
    pybabel init -i locales/{domain_name}.pot -d locales -l uz_Latn
Now you can translate the texts located in locales/{language}/LC_MESSAGES/{domain_name}.po
After you translated all the texts you need to compile .po files:
    pybabel compile -d locales
When you delete/update your texts you also need to update them in .po files:
    pybabel extract -o locales/{domain_name}.pot -k __:1,2 --input-dirs .
    pybabel update -i locales/{domain_name}.pot -d locales
    - translate
    pybabel compile -d locales
If you have any exceptions check:
    - you have installed babel
    - translations are ready, so you just compiled it
    - in the commands above you replaced {domain_name} to messages
    - you are writing commands from correct path in terminal
"""

import keyboards
from i18n_class import I18N

from telebot import TeleBot, apihelper, custom_filters, types
from telebot.storage.memory_storage import StateMemoryStorage

apihelper.ENABLE_MIDDLEWARE = True
storage = StateMemoryStorage()
# IMPORTANT! This example works only if polling is non-threaded.
bot = TeleBot("", state_storage=storage, threaded=False)

i18n = I18N(translations_path="locales", domain_name="messages")
_ = i18n.gettext  # for singular translations
__ = i18n.ngettext  # for plural translations

# These are example storages, do not use it in a production development
users_lang = {}
users_clicks = {}


@bot.middleware_handler(update_types=["message", "callback_query"])
def set_contex_language(bot_instance, message):
    i18n.context_lang.language = users_lang.get(message.from_user.id, "en")


@bot.message_handler(commands=["start"])
def start_handler(message: types.Message):
    text = _(
        "Hello, {user_fist_name}!\n"
        "This is the example of multilanguage bot.\n"
        "Available commands:\n\n"
        "/lang - change your language\n"
        "/plural - pluralization example"
    )

    # remember don't use f string for interpolation, use .format method instead
    text = text.format(user_fist_name=message.from_user.first_name)
    bot.send_message(message.from_user.id, text)


@bot.message_handler(commands=["lang"])
def change_language_handler(message: types.Message):
    bot.send_message(
        message.chat.id,
        "Choose language\nВыберите язык\nTilni tanlang",
        reply_markup=keyboards.languages_keyboard(),
    )


@bot.callback_query_handler(
    func=None, text=custom_filters.TextFilter(contains=["en", "ru", "uz_Latn"])
)
def language_handler(call: types.CallbackQuery):
    lang = call.data
    users_lang[call.from_user.id] = lang

    # When you change user's language, pass language explicitly coz it's not changed in context
    bot.edit_message_text(
        _("Language has been changed", lang=lang), call.from_user.id, call.message.id
    )
    bot.delete_state(call.from_user.id)


@bot.message_handler(commands=["plural"])
def pluralization_handler(message: types.Message):
    if not users_clicks.get(message.from_user.id):
        users_clicks[message.from_user.id] = 0
    clicks = users_clicks[message.from_user.id]

    text = __(
        singular="You have {number} click",
        plural="You have {number} clicks",
        n=clicks,
    )
    text = _("This is clicker.\n\n") + text.format(number=clicks)
    bot.send_message(message.chat.id, text, reply_markup=keyboards.clicker_keyboard(_))


@bot.callback_query_handler(func=None, text=custom_filters.TextFilter(equals="click"))
def click_handler(call: types.CallbackQuery):
    if not users_clicks.get(call.from_user.id):
        users_clicks[call.from_user.id] = 1
    else:
        users_clicks[call.from_user.id] += 1

    clicks = users_clicks[call.from_user.id]

    text = __(
        singular="You have {number} click", plural="You have {number} clicks", n=clicks
    )
    text = _("This is clicker.\n\n") + text.format(number=clicks)
    bot.edit_message_text(
        text,
        call.from_user.id,
        call.message.message_id,
        reply_markup=keyboards.clicker_keyboard(_),
    )


if __name__ == "__main__":
    bot.add_custom_filter(custom_filters.TextMatchFilter())
    bot.infinity_polling()



"""
Callback data factory's file.
"""

"""
Copyright (c) 2017-2018 Alex Root Junior

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so, subject to the
following conditions:

The above copyright notice and this permission notice shall be included in all copies
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
OR OTHER DEALINGS IN THE SOFTWARE.


This file was added during the pull request. The maintainers overlooked that it was copied
"as is" from another project and they do not consider it as a right way to develop a project.
However, due to backward compatibility we had to leave this file in the project with the above
copyright added, as it is required by the original project license.
"""

import typing


class CallbackDataFilter:
    """
    Filter for CallbackData.
    """

    def __init__(self, factory, config: typing.Dict[str, str]):
        self.config = config
        self.factory = factory

    def check(self, query) -> bool:
        """
        Checks if query.data appropriates to specified config

        :param query: telebot.types.CallbackQuery
        :type query: telebot.types.CallbackQuery

        :return: True if query.data appropriates to specified config
        :rtype: bool
        """

        try:
            data = self.factory.parse(query.data)
        except ValueError:
            return False

        for key, value in self.config.items():
            if isinstance(value, (list, tuple, set, frozenset)):
                if data.get(key) not in value:
                    return False
            elif data.get(key) != value:
                return False
        return True


class CallbackData:
    """
    Callback data factory
    This class will help you to work with CallbackQuery
    """

    def __init__(self, *parts, prefix: str, sep=":"):
        if not isinstance(prefix, str):
            raise TypeError(
                f"Prefix must be instance of str not {type(prefix).__name__}"
            )
        if not prefix:
            raise ValueError("Prefix can't be empty")
        if sep in prefix:
            raise ValueError(f"Separator {sep!r} can't be used in prefix")

        self.prefix = prefix
        self.sep = sep

        self._part_names = parts

    def new(self, *args, **kwargs) -> str:
        """
        Generate callback data

        :param args: positional parameters of CallbackData instance parts
        :param kwargs: named parameters
        :return: str
        """
        args = list(args)

        data = [self.prefix]

        for part in self._part_names:
            value = kwargs.pop(part, None)
            if value is None:
                if args:
                    value = args.pop(0)
                else:
                    raise ValueError(f"Value for {part!r} was not passed!")

            if value is not None and not isinstance(value, str):
                value = str(value)

            if self.sep in value:
                raise ValueError(
                    f"Symbol {self.sep!r} is defined as the separator and can't be used in parts' values"
                )

            data.append(value)

        if args or kwargs:
            raise TypeError("Too many arguments were passed!")

        callback_data = self.sep.join(data)

        if len(callback_data.encode()) > 64:
            raise ValueError("Resulted callback data is too long!")

        return callback_data

    def parse(self, callback_data: str) -> typing.Dict[str, str]:
        """
        Parse data from the callback data

        :param callback_data: string, use to telebot.types.CallbackQuery to parse it from string to a dict
        :return: dict parsed from callback data
        """

        prefix, *parts = callback_data.split(self.sep)
        if prefix != self.prefix:
            raise ValueError("Passed callback data can't be parsed with that prefix.")
        elif len(parts) != len(self._part_names):
            raise ValueError("Invalid parts count!")

        result = {"@": prefix}
        result.update(zip(self._part_names, parts))
        return result

    def filter(self, **config) -> CallbackDataFilter:
        """
        Generate filter

        :param config: specified named parameters will be checked with CallbackQuery.data
        :return: CallbackDataFilter class
        """

        for key in config.keys():
            if key not in self._part_names:
                raise ValueError(f"Invalid field name {key!r}")
        return CallbackDataFilter(self, config)



from telebot import TeleBot, types


def hello_handler(message: types.Message, bot: TeleBot):
    bot.send_message(message.chat.id, "Hi :)")


def echo_handler(message: types.Message, bot: TeleBot):
    bot.send_message(message.chat.id, message.text)


def register_handlers(bot: TeleBot):
    bot.register_message_handler(
        hello_handler, func=lambda message: message.text == "Hello", pass_bot=True
    )
    bot.register_message_handler(echo_handler, pass_bot=True)



# -*- coding: utf-8 -*-
import hmac
import logging

# noinspection PyPep8Naming
import queue as Queue
import re
import threading
import traceback
from hashlib import sha256
from typing import Any, Callable, Dict, List, Optional, Union
from urllib.parse import parse_qsl

from telebot import types
from telebot.service_utils import (
    chunks,
    generate_random_token,
    is_bytes,
    is_dict,
    is_pil_image,
    is_string,
    pil_image_to_file,
)

try:
    import ujson as json
except ImportError:
    import json

MAX_MESSAGE_LENGTH = 4096

logger = logging.getLogger("TeleBot")

thread_local = threading.local()

#: Contains all media content types.
content_type_media = [
    "text",
    "animation",
    "audio",
    "document",
    "photo",
    "sticker",
    "story",
    "video",
    "video_note",
    "voice",
    "contact",
    "dice",
    "game",
    "poll",
    "venue",
    "location",
    "invoice",
    "successful_payment",
    "connected_website",
    "passport_data",
    "web_app_data",
]

#: Contains all service content types such as `User joined the group`.
content_type_service = [
    "new_chat_members",
    "left_chat_member",
    "new_chat_title",
    "new_chat_photo",
    "delete_chat_photo",
    "group_chat_created",
    "supergroup_chat_created",
    "channel_chat_created",
    "message_auto_delete_timer_changed",
    "migrate_to_chat_id",
    "migrate_from_chat_id",
    "pinned_message",
    "users_shared",
    "chat_shared",
    "write_access_allowed",
    "proximity_alert_triggered",
    "forum_topic_created",
    "forum_topic_edited",
    "forum_topic_closed",
    "forum_topic_reopened",
    "general_forum_topic_hidden",
    "general_forum_topic_unhidden",
    "giveaway_created",
    "giveaway",
    "giveaway_winners",
    "giveaway_completed",
    "video_chat_scheduled",
    "video_chat_started",
    "video_chat_ended",
    "video_chat_participants_invited",
]

#: All update types, should be used for allowed_updates parameter in polling.
update_types = [
    "message",
    "edited_message",
    "channel_post",
    "edited_channel_post",
    "inline_query",
    "chosen_inline_result",
    "callback_query",
    "shipping_query",
    "pre_checkout_query",
    "poll",
    "poll_answer",
    "my_chat_member",
    "chat_member",
    "chat_join_request",
    "message_reaction",
    "message_reaction_count",
    "chat_boost",
    "removed_chat_boost",
    "business_connection",
    "business_message",
    "edited_business_message",
    "deleted_business_messages",
]


class WorkerThread(threading.Thread):
    """
    :meta private:
    """

    count = 0

    def __init__(self, exception_callback=None, queue=None, name=None):
        if not name:
            name = "WorkerThread{0}".format(self.__class__.count + 1)
            self.__class__.count += 1
        if not queue:
            queue = Queue.Queue()

        threading.Thread.__init__(self, name=name)
        self.queue = queue
        self.daemon = True

        self.received_task_event = threading.Event()
        self.done_event = threading.Event()
        self.exception_event = threading.Event()
        self.continue_event = threading.Event()

        self.exception_callback = exception_callback
        self.exception_info = None
        self._running = True
        self.start()

    def run(self):
        while self._running:
            try:
                task, args, kwargs = self.queue.get(block=True, timeout=0.5)
                self.continue_event.clear()
                self.received_task_event.clear()
                self.done_event.clear()
                self.exception_event.clear()
                logger.debug("Received task")
                self.received_task_event.set()

                task(*args, **kwargs)
                logger.debug("Task complete")
                self.done_event.set()
            except Queue.Empty:
                pass
            except Exception as e:
                logger.debug(
                    type(e).__name__
                    + " occurred, args="
                    + str(e.args)
                    + "\n"
                    + traceback.format_exc()
                )
                self.exception_info = e
                self.exception_event.set()
                if self.exception_callback:
                    self.exception_callback(self, self.exception_info)
                self.continue_event.wait()

    def put(self, task, *args, **kwargs):
        self.queue.put((task, args, kwargs))

    def raise_exceptions(self):
        if self.exception_event.is_set():
            raise self.exception_info

    def clear_exceptions(self):
        self.exception_event.clear()
        self.continue_event.set()

    def stop(self):
        self._running = False


class ThreadPool:
    """
    :meta private:
    """

    def __init__(self, telebot, num_threads=2):
        self.telebot = telebot
        self.tasks = Queue.Queue()
        self.workers = [
            WorkerThread(self.on_exception, self.tasks) for _ in range(num_threads)
        ]
        self.num_threads = num_threads

        self.exception_event = threading.Event()
        self.exception_info = None

    def put(self, func, *args, **kwargs):
        self.tasks.put((func, args, kwargs))

    def on_exception(self, worker_thread, exc_info):
        if self.telebot.exception_handler is not None:
            handled = self.telebot.exception_handler.handle(exc_info)
        else:
            handled = False
        if not handled:
            self.exception_info = exc_info
            self.exception_event.set()
        worker_thread.continue_event.set()

    def raise_exceptions(self):
        if self.exception_event.is_set():
            raise self.exception_info

    def clear_exceptions(self):
        self.exception_event.clear()

    def close(self):
        for worker in self.workers:
            worker.stop()
        for worker in self.workers:
            if worker != threading.current_thread():
                worker.join()


class AsyncTask:
    """
    :meta private:
    """

    def __init__(self, target, *args, **kwargs):
        self.target = target
        self.args = args
        self.kwargs = kwargs

        self.done = False
        self.thread = threading.Thread(target=self._run)
        self.thread.start()

    def _run(self):
        try:
            self.result = self.target(*self.args, **self.kwargs)
        except Exception as e:
            self.result = e
        self.done = True

    def wait(self):
        if not self.done:
            self.thread.join()
        if isinstance(self.result, BaseException):
            raise self.result
        else:
            return self.result


class CustomRequestResponse:
    """
    :meta private:
    """

    def __init__(self, json_text, status_code=200, reason=""):
        self.status_code = status_code
        self.text = json_text
        self.reason = reason

    def json(self):
        return json.loads(self.text)


def async_dec():
    """
    :meta private:
    """

    def decorator(fn):
        def wrapper(*args, **kwargs):
            return AsyncTask(fn, *args, **kwargs)

        return wrapper

    return decorator


def is_command(text: str) -> bool:
    r"""
    Checks if `text` is a command. Telegram chat commands start with the '/' character.

    :param text: Text to check.
    :type text: :obj:`str`

    :return: True if `text` is a command, else False.
    :rtype: :obj:`bool`
    """
    if text is None:
        return False
    return text.startswith("/")


def extract_command(text: str) -> Union[str, None]:
    """
    Extracts the command from `text` (minus the '/') if `text` is a command (see is_command).
    If `text` is not a command, this function returns None.

    .. code-block:: python3
        :caption: Examples:

        extract_command('/help'): 'help'
        extract_command('/help@BotName'): 'help'
        extract_command('/search black eyed peas'): 'search'
        extract_command('Good day to you'): None

    :param text: String to extract the command from
    :type text: :obj:`str`

    :return: the command if `text` is a command (according to is_command), else None.
    :rtype: :obj:`str` or :obj:`None`
    """
    if text is None:
        return None
    return text.split()[0].split("@")[0][1:] if is_command(text) else None


def extract_arguments(text: str) -> str or None:
    """
    Returns the argument after the command.

    .. code-block:: python3
        :caption: Examples:

        extract_arguments("/get name"): 'name'
        extract_arguments("/get"): ''
        extract_arguments("/get@botName name"): 'name'

    :param text: String to extract the arguments from a command
    :type text: :obj:`str`

    :return: the arguments if `text` is a command (according to is_command), else None.
    :rtype: :obj:`str` or :obj:`None`
    """
    regexp = re.compile(r"/\w*(@\w*)*\s*([\s\S]*)", re.IGNORECASE)
    result = regexp.match(text)
    return result.group(2) if is_command(text) else None


def extract_entity(text: str, e: types.MessageEntity) -> str:
    """
    Returns the content of the entity.

    :param text: The text of the message the entity belongs to
    :type text: :obj:`str`

    :param e: The entity to extract
    :type e: :obj:`MessageEntity`

    :return: The content of the entity
    :rtype: :obj:`str`
    """
    offset = 0
    start = 0
    encoded_text = text.encode()
    end = len(encoded_text)
    i = 0

    for byte in encoded_text:
        if (byte & 0xC0) != 0x80:
            if offset == e.offset:
                start = i
            elif offset - e.offset == e.length:
                end = i
                break
            if byte >= 0xF0:
                offset += 2
            else:
                offset += 1
        i += 1

    return encoded_text[start:end].decode()


def split_string(text: str, chars_per_string: int) -> List[str]:
    """
    Splits one string into multiple strings, with a maximum amount of `chars_per_string` characters per string.
    This is very useful for splitting one giant message into multiples.

    :param text: The text to split
    :type text: :obj:`str`

    :param chars_per_string: The number of characters per line the text is split into.
    :type chars_per_string: :obj:`int`

    :return: The splitted text as a list of strings.
    :rtype: :obj:`list` of :obj:`str`
    """
    return [
        text[i : i + chars_per_string] for i in range(0, len(text), chars_per_string)
    ]


def smart_split(text: str, chars_per_string: int = MAX_MESSAGE_LENGTH) -> List[str]:
    r"""
    Splits one string into multiple strings, with a maximum amount of `chars_per_string` characters per string.
    This is very useful for splitting one giant message into multiples.
    If `chars_per_string` > 4096: `chars_per_string` = 4096.
    Splits by '\n', '. ' or ' ' in exactly this priority.

    :param text: The text to split
    :type text: :obj:`str`

    :param chars_per_string: The number of maximum characters per part the text is split to.
    :type chars_per_string: :obj:`int`

    :return: The splitted text as a list of strings.
    :rtype: :obj:`list` of :obj:`str`
    """

    def _text_before_last(substr: str) -> str:
        return substr.join(part.split(substr)[:-1]) + substr

    if chars_per_string > MAX_MESSAGE_LENGTH:
        chars_per_string = MAX_MESSAGE_LENGTH

    parts = []
    while True:
        if len(text) < chars_per_string:
            parts.append(text)
            return parts

        part = text[:chars_per_string]

        if "\n" in part:
            part = _text_before_last("\n")
        elif ". " in part:
            part = _text_before_last(". ")
        elif " " in part:
            part = _text_before_last(" ")

        parts.append(part)
        text = text[len(part) :]


def escape(text: str) -> Optional[str]:
    """
    Replaces the following chars in `text` ('&' with '&amp;', '<' with '&lt;' and '>' with '&gt;').

    :param text: the text to escape
    :return: the escaped text
    """
    chars = {"&": "&amp;", "<": "&lt;", ">": "&gt;"}
    if text is None:
        return None
    for old, new in chars.items():
        text = text.replace(old, new)
    return text


def user_link(user: types.User, include_id: bool = False) -> str:
    """
    Returns an HTML user link. This is useful for reports.
    Attention: Don't forget to set parse_mode to 'HTML'!


    .. code-block:: python3
        :caption: Example:

        bot.send_message(your_user_id, user_link(message.from_user) + ' started the bot!', parse_mode='HTML')

    .. note::
        You can use formatting.* for all other formatting options(bold, italic, links, and etc.)
        This method is kept for backward compatibility, and it is recommended to use formatting.* for
        more options.

    :param user: the user (not the user_id)
    :type user: :obj:`telebot.types.User`

    :param include_id: include the user_id
    :type include_id: :obj:`bool`

    :return: HTML user link
    :rtype: :obj:`str`
    """
    name = escape(user.first_name)
    return f"<a href='tg://user?id={user.id}'>{name}</a>" + (
        f" (<pre>{user.id}</pre>)" if include_id else ""
    )


def quick_markup(
    values: Dict[str, Dict[str, Any]], row_width: int = 2
) -> types.InlineKeyboardMarkup:
    """
    Returns a reply markup from a dict in this format: {'text': kwargs}
    This is useful to avoid always typing 'btn1 = InlineKeyboardButton(...)' 'btn2 = InlineKeyboardButton(...)'

    Example:

    .. code-block:: python3
        :caption: Using quick_markup:

        from telebot.util import quick_markup

        markup = quick_markup({
            'Twitter': {'url': 'https://twitter.com'},
            'Facebook': {'url': 'https://facebook.com'},
            'Back': {'callback_data': 'whatever'}
        }, row_width=2)
        # returns an InlineKeyboardMarkup with two buttons in a row, one leading to Twitter, the other to facebook
        # and a back button below

        # kwargs can be:
        {
            'url': None,
            'callback_data': None,
            'switch_inline_query': None,
            'switch_inline_query_current_chat': None,
            'callback_game': None,
            'pay': None,
            'login_url': None,
            'web_app': None
        }

    :param values: a dict containing all buttons to create in this format: {text: kwargs} {str:}
    :type values: :obj:`dict`

    :param row_width: number of :class:`telebot.types.InlineKeyboardButton` objects on each row
    :type row_width: :obj:`int`

    :return: InlineKeyboardMarkup
    :rtype: :obj:`types.InlineKeyboardMarkup`
    """
    markup = types.InlineKeyboardMarkup(row_width=row_width)
    buttons = [
        types.InlineKeyboardButton(text=text, **kwargs)
        for text, kwargs in values.items()
    ]
    markup.add(*buttons)
    return markup


# CREDITS TO http://stackoverflow.com/questions/12317940#answer-12320352
def or_set(self):
    """
    :meta private:
    """
    self._set()
    self.changed()


def or_clear(self):
    """
    :meta private:
    """
    self._clear()
    self.changed()


def orify(e, changed_callback):
    """
    :meta private:
    """
    if not hasattr(e, "_set"):
        e._set = e.set
    if not hasattr(e, "_clear"):
        e._clear = e.clear
    e.changed = changed_callback
    e.set = lambda: or_set(e)
    e.clear = lambda: or_clear(e)


def OrEvent(*events):
    """
    :meta private:
    """
    or_event = threading.Event()

    def changed():
        bools = [ev.is_set() for ev in events]
        if any(bools):
            or_event.set()
        else:
            or_event.clear()

    def busy_wait():
        while not or_event.is_set():
            # noinspection PyProtectedMember
            or_event._wait(3)

    for e in events:
        orify(e, changed)
    or_event._wait = or_event.wait
    or_event.wait = busy_wait
    changed()
    return or_event


def per_thread(key, construct_value, reset=False):
    """
    :meta private:
    """
    if reset or not hasattr(thread_local, key):
        value = construct_value()
        setattr(thread_local, key, value)

    return getattr(thread_local, key)


def deprecated(
    warn: bool = True, alternative: Optional[Callable] = None, deprecation_text=None
):
    """
    Use this decorator to mark functions as deprecated.
    When the function is used, an info (or warning if `warn` is True) is logged.

    :meta private:

    :param warn: If True a warning is logged else an info
    :type warn: :obj:`bool`

    :param alternative: The new function to use instead
    :type alternative: :obj:`Callable`

    :param deprecation_text: Custom deprecation text
    :type deprecation_text: :obj:`str`

    :return: The decorated function
    """

    def decorator(function):
        def wrapper(*args, **kwargs):
            info = f"`{function.__name__}` is deprecated."
            if alternative:
                info += f" Use `{alternative.__name__}` instead"
            if deprecation_text:
                info += " " + deprecation_text
            if not warn:
                logger.info(info)
            else:
                logger.warning(info)
            return function(*args, **kwargs)

        return wrapper

    return decorator


# Cloud helpers
def webhook_google_functions(bot, request):
    """
    A webhook endpoint for Google Cloud Functions FaaS.

    :param bot: The bot instance
    :type bot: :obj:`telebot.TeleBot` or :obj:`telebot.async_telebot.AsyncTeleBot`

    :param request: The request object
    :type request: :obj:`flask.Request`

    :return: The response object
    """
    if request.is_json:
        try:
            request_json = request.get_json()
            update = types.Update.de_json(request_json)
            bot.process_new_updates([update])
            return ""
        except Exception as e:
            print(e)
            return "Bot FAIL", 400
    else:
        return "Bot ON"


def antiflood(function: Callable, *args, number_retries=5, **kwargs):
    """
    Use this function inside loops in order to avoid getting TooManyRequests error.
    Example:

    .. code-block:: python3

        from telebot.util import antiflood
        for chat_id in chat_id_list:
        msg = antiflood(bot.send_message, chat_id, text)

    :param function: The function to call
    :type function: :obj:`Callable`

    :param number_retries: Number of retries to send
    :type function: :obj:int

    :param args: The arguments to pass to the function
    :type args: :obj:`tuple`

    :param kwargs: The keyword arguments to pass to the function
    :type kwargs: :obj:`dict`

    :return: None
    """
    from time import sleep

    from telebot.apihelper import ApiTelegramException

    for _ in range(number_retries - 1):
        try:
            return function(*args, **kwargs)
        except ApiTelegramException as ex:
            if ex.error_code == 429:
                sleep(ex.result_json["parameters"]["retry_after"])
            else:
                raise
    else:
        return function(*args, **kwargs)


def parse_web_app_data(token: str, raw_init_data: str):
    """
    Parses web app data.

    :param token: The bot token
    :type token: :obj:`str`

    :param raw_init_data: The raw init data
    :type raw_init_data: :obj:`str`

    :return: The parsed init data
    """
    is_valid = validate_web_app_data(token, raw_init_data)
    if not is_valid:
        return False

    result = {}
    for key, value in parse_qsl(raw_init_data):
        try:
            value = json.loads(value)
        except json.JSONDecodeError:
            result[key] = value
        else:
            result[key] = value
    return result


def validate_web_app_data(token: str, raw_init_data: str):
    """
    Validates web app data.

    :param token: The bot token
    :type token: :obj:`str`

    :param raw_init_data: The raw init data
    :type raw_init_data: :obj:`str`

    :return: The parsed init data
    """
    try:
        parsed_data = dict(parse_qsl(raw_init_data))
    except ValueError:
        return False
    if "hash" not in parsed_data:
        return False

    init_data_hash = parsed_data.pop("hash")
    data_check_string = "\n".join(
        f"{key}={value}" for key, value in sorted(parsed_data.items())
    )
    secret_key = hmac.new(key=b"WebAppData", msg=token.encode(), digestmod=sha256)

    return (
        hmac.new(secret_key.digest(), data_check_string.encode(), sha256).hexdigest()
        == init_data_hash
    )


__all__ = (
    "content_type_media",
    "content_type_service",
    "update_types",
    "WorkerThread",
    "AsyncTask",
    "CustomRequestResponse",
    "async_dec",
    "deprecated",
    "is_bytes",
    "is_string",
    "is_dict",
    "is_pil_image",
    "chunks",
    "generate_random_token",
    "pil_image_to_file",
    "is_command",
    "extract_command",
    "extract_arguments",
    "split_string",
    "smart_split",
    "escape",
    "user_link",
    "quick_markup",
    "antiflood",
    "parse_web_app_data",
    "validate_web_app_data",
    "or_set",
    "or_clear",
    "orify",
    "OrEvent",
    "per_thread",
    "webhook_google_functions",
)



from time import sleep, time

import telebot
from telebot.types import (  # Only for creating Inline Buttons, not necessary for creating Invite Links
    InlineKeyboardButton,
    InlineKeyboardMarkup,
)

Token = "api_token"  # Your Bot Access Token
Group_ID = -1234567890  # Group ID for which invite link is to be created

bot = telebot.TeleBot(Token, parse_mode="HTML")


# /start command message
@bot.message_handler(commands=["start"])
def startmsg(msg):
    bot.reply_to(msg, "Hey there, I'm a bot made by pyTelegramBotAPI!")


# Get notified of incoming members in group
@bot.message_handler(content_types=["new_chat_members"])
def newmember(msg):
    # Create an invite link class that contains info about the created invite link using create_chat_invite_link() with parameters
    invite = bot.create_chat_invite_link(
        Group_ID, member_limit=1, expire_date=int(time()) + 45
    )  # Here, the link will auto-expire in 45 seconds
    InviteLink = invite.invite_link  # Get the actual invite link from 'invite' class

    mrkplink = InlineKeyboardMarkup()  # Created Inline Keyboard Markup
    mrkplink.add(
        InlineKeyboardButton("Join our group 🚀", url=InviteLink)
    )  # Added Invite Link to Inline Keyboard

    bot.send_message(
        msg.chat.id,
        f"Hey there {msg.from_user.first_name}, Click the link below to join our Official Group.",
        reply_markup=mrkplink,
    )

    # This will send a message with the newly-created invite link as markup button.
    # The member limit will be 1 and expiring time will be 45 sec.


bot.infinity_polling()



import telebot
from telebot.async_telebot import AsyncTeleBot

bot = AsyncTeleBot("TOKEN")


@bot.chat_join_request_handler()
async def make_some(message: telebot.types.ChatJoinRequest):
    await bot.send_message(message.chat.id, "I accepted a new user!")
    await bot.approve_chat_join_request(message.chat.id, message.from_user.id)


import asyncio

asyncio.run(bot.polling())



#!/usr/bin/python

# This is a simple echo bot using the decorator mechanism.
# It echoes any incoming text messages.
import asyncio

from telebot.async_telebot import AsyncTeleBot

bot = AsyncTeleBot("TOKEN")


# Handle '/start' and '/help'
@bot.message_handler(commands=["help", "start"])
async def send_welcome(message):
    text = "Hi, I am EchoBot.\nJust write me something and I will repeat it!"
    await bot.reply_to(message, text)


# Handle all other messages with content_type 'text' (content_types defaults to ['text'])
@bot.message_handler(func=lambda message: True)
async def echo_message(message):
    await bot.reply_to(message, message.text)


asyncio.run(bot.polling())



#!/usr/bin/python

# This is a set_my_commands example.
# Press on [/] button in telegram client.
# Important, to update the command menu, be sure to exit the chat with the bot and enter to chat again
# Important, command for chat_id and for group have a higher priority than for all

import telebot

API_TOKEN = "<api_token>"
bot = telebot.TeleBot(API_TOKEN)

# use in for delete with the necessary scope and language_code if necessary
bot.delete_my_commands(scope=None, language_code=None)

bot.set_my_commands(
    commands=[
        telebot.types.BotCommand("command1", "command1 description"),
        telebot.types.BotCommand("command2", "command2 description"),
    ],
    # scope=telebot.types.BotCommandScopeChat(12345678)  # use for personal command for users
    # scope=telebot.types.BotCommandScopeAllPrivateChats()  # use for all private chats
)

# check command
cmd = bot.get_my_commands(scope=None, language_code=None)
print([c.to_json() for c in cmd])



#!/usr/bin/python

# This example shows how to implement i18n (internationalization) l10n (localization) to create
# multi-language bots with middleware handler.
#
# Also, you could check language code in handler itself too.
# But this example just to show the work of middlewares.

import logging

import telebot
from telebot import asyncio_handler_backends
from telebot.async_telebot import AsyncTeleBot

logger = telebot.logger
telebot.logger.setLevel(logging.DEBUG)  # Outputs debug messages to console.

TRANSLATIONS = {"hello": {"en": "hello", "ru": "привет", "uz": "salom"}}


bot = AsyncTeleBot("TOKEN")


class LanguageMiddleware(asyncio_handler_backends.BaseMiddleware):
    def __init__(self):
        self.update_types = [
            "message"
        ]  # Update types that will be handled by this middleware.

    async def pre_process(self, message, data):
        data["response"] = TRANSLATIONS["hello"][message.from_user.language_code]

    async def post_process(self, message, data, exception):
        if exception:  # You can get exception occured in handler.
            logger.exception(str(exception))


bot.setup_middleware(LanguageMiddleware())  # do not forget to setup


@bot.message_handler(commands=["start"])
async def start(message, data: dict):
    # you can get the data in handler too.
    # Not necessary to create data parameter in handler function.
    await bot.send_message(message.chat.id, data["response"])


import asyncio

asyncio.run(bot.polling())



import telebot
from telebot.async_telebot import AsyncTeleBot

bot = AsyncTeleBot("TOKEN")


@bot.message_handler(commands=["photo"])
async def photo_send(message: telebot.types.Message):
    with open("test.png", "rb") as new_file:
        await bot.send_photo(message.chat.id, new_file)


@bot.message_handler(commands=["document"])
async def document_send(message: telebot.types.Message):
    with open("test.docx", "rb") as new_file:
        await bot.send_document(message.chat.id, new_file)


@bot.message_handler(commands=["photos"])
async def photos_send(message: telebot.types.Message):
    with open("test.png", "rb") as new_file, open("test2.png", "rb") as new_file2:
        await bot.send_media_group(
            message.chat.id,
            [
                telebot.types.InputMediaPhoto(new_file),
                telebot.types.InputMediaPhoto(new_file2),
            ],
        )


import asyncio

asyncio.run(bot.polling())



import asyncio  # for future uses
import ssl

import aiohttp
import certifi

from telebot import types

try:
    import ujson as json
except ImportError:
    import json

import os

API_URL = "https://api.telegram.org/bot{0}/{1}"

import logging
from datetime import datetime

from telebot import util

logger = logging.getLogger("TeleBot")

proxy = None
session = None

FILE_URL = None

REQUEST_TIMEOUT = 300
MAX_RETRIES = 3

REQUEST_LIMIT = 50


class SessionManager:
    def __init__(self) -> None:
        self.session = None
        self.ssl_context = ssl.create_default_context(cafile=certifi.where())

    async def create_session(self):
        self.session = aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(
                limit=REQUEST_LIMIT, ssl_context=self.ssl_context
            )
        )
        return self.session

    async def get_session(self):
        if self.session is None:
            self.session = await self.create_session()
            return self.session

        if self.session.closed:
            self.session = await self.create_session()

        # noinspection PyProtectedMember
        if not self.session._loop.is_running():
            await self.session.close()
            self.session = await self.create_session()
        return self.session


session_manager = SessionManager()
session_manager_bale = SessionManager()


async def _process_request(token, url, method="get", params=None, files=None, **kwargs):
    # Let's resolve all timeout parameters.
    # getUpdates parameter may contain 2 parameters: request_timeout & timeout.
    # other methods may contain timeout parameter that should be applied to
    # ClientTimeout only.
    # timeout should be added to params for getUpdates. All other timeout's should be used
    # for request timeout.
    # here we got request_timeout, so this is getUpdates method.
    if "request_timeout" in kwargs:
        request_timeout = kwargs.pop("request_timeout")

    else:
        # let's check for timeout in params
        request_timeout = params.pop("timeout", None) if params else None
        # we will apply default request_timeout if there is no timeout in params
        # otherwise, we will use timeout parameter applied for payload.

    request_timeout = REQUEST_TIMEOUT if request_timeout is None else request_timeout

    # Preparing data by adding all parameters and files to FormData
    file_id = params.get("file_id", None)
    params = _prepare_data(params, files)

    timeout = aiohttp.ClientTimeout(total=request_timeout)
    got_result = False
    current_try = 0
    if len(token) == 51:
        session = await session_manager_bale.get_session()
    else:
        session = await session_manager.get_session()
    while not got_result and current_try < MAX_RETRIES - 1:
        current_try += 1
        try:
            if len(token) == 51:
                if url == "getFile":
                    return {
                        "file_id": file_id,
                        "file_path": file_id,
                        "file_unique_id": file_id,
                    }
                API_URL = "https://tapi.bale.ai/bot{0}/{1}"
            else:
                API_URL = "https://api.telegram.org/bot{0}/{1}"
            async with session.request(
                method=method,
                url=API_URL.format(token, url),
                data=params,
                timeout=timeout,
                proxy=proxy,
            ) as resp:
                got_result = True
                logger.debug(
                    "Request: method={0} url={1} params={2} files={3} request_timeout={4} current_try={5}".format(
                        method, url, params, files, request_timeout, current_try
                    ).replace(
                        token, token.split(":")[0] + ":{TOKEN}"
                    )
                )

                json_result = await _check_result(url, resp)
                if json_result:
                    return json_result["result"]
        except (ApiTelegramException, ApiInvalidJSONException, ApiHTTPException) as e:
            raise e
        except aiohttp.ClientError as e:
            logger.error("Aiohttp ClientError: {0}".format(e.__class__.__name__))
        except Exception as e:
            logger.error(f"Unknown error: {e.__class__.__name__}")
        if not got_result:
            raise RequestTimeout(
                "Request timeout. Request: method={0} url={1} params={2} files={3} request_timeout={4}".format(
                    method, url, params, files, request_timeout, current_try
                )
            )


def _prepare_file(obj):
    """
    Prepares file for upload.
    """
    name = getattr(obj, "name", None)
    if name and isinstance(name, str) and name[0] != "<" and name[-1] != ">":
        return os.path.basename(name)


def _prepare_data(params=None, files=None):
    """
    Adds the parameters and files to the request.

    :param params:
    :param files:
    :return:
    """
    data = aiohttp.formdata.FormData(quote_fields=False)

    if params:
        for key, value in params.items():
            data.add_field(key, str(value))
    if files:
        for key, f in files.items():
            if isinstance(f, tuple):
                if len(f) == 2:
                    file_name, file = f
                else:
                    raise ValueError(
                        "Tuple must have exactly 2 elements: filename, fileobj"
                    )
            elif isinstance(f, types.InputFile):
                file_name = f.file_name
                file = f.file
            else:
                file_name, file = _prepare_file(f) or key, f

            data.add_field(key, file, filename=file_name)

    return data


async def _convert_markup(markup):
    if isinstance(markup, types.JsonSerializable):
        return markup.to_json()
    return markup


async def get_me(token):
    method_url = r"getMe"
    return await _process_request(token, method_url)


async def log_out(token):
    method_url = r"logOut"
    return await _process_request(token, method_url)


async def close(token):
    method_url = r"close"
    return await _process_request(token, method_url)


async def get_file(token, file_id):
    method_url = r"getFile"
    return await _process_request(token, method_url, params={"file_id": file_id})


async def get_file_url(token, file_id):
    if FILE_URL is None:
        if len(token) == 51:
            return f"https://tapi.bale.ai/file/bot{token}/{file_id}"
        return "https://api.telegram.org/file/bot{0}/{1}".format(
            token, (await get_file(token, file_id))["file_path"]
        )
    else:
        # noinspection PyUnresolvedReferences
        return FILE_URL.format(token, (await get_file(token, file_id))["file_path"])


async def download_file(token, file_path):
    if FILE_URL is None:
        if len(token) == 51:
            url = f"https://tapi.bale.ai/file/bot{token}/{file_path}"
        else:
            url = "https://api.telegram.org/file/bot{0}/{1}".format(token, file_path)
    if len(token) == 51:
        session = await session_manager_bale.get_session()
    else:
        session = await session_manager.get_session()

    async with session.get(url, proxy=proxy) as response:
        if response.status != 200:
            raise ApiHTTPException("Download file", response)
        result = await response.read()

    return result


async def set_webhook(
    token,
    url=None,
    certificate=None,
    max_connections=None,
    allowed_updates=None,
    ip_address=None,
    drop_pending_updates=None,
    timeout=None,
    secret_token=None,
):
    method_url = r"setWebhook"
    payload = {
        "url": url if url else "",
    }
    files = None
    if certificate:
        files = {"certificate": certificate}
    if max_connections:
        payload["max_connections"] = max_connections
    if allowed_updates is not None:  # Empty lists should pass
        payload["allowed_updates"] = json.dumps(allowed_updates)
    if ip_address is not None:  # Empty string should pass
        payload["ip_address"] = ip_address
    if drop_pending_updates is not None:  # Any bool value should pass
        payload["drop_pending_updates"] = drop_pending_updates
    if timeout:
        payload["timeout"] = timeout
    if secret_token:
        payload["secret_token"] = secret_token
    return await _process_request(token, method_url, params=payload, files=files)


async def delete_webhook(token, drop_pending_updates=None, timeout=None):
    method_url = r"deleteWebhook"
    payload = {}
    if drop_pending_updates is not None:  # Any bool value should pass
        payload["drop_pending_updates"] = drop_pending_updates
    if timeout:
        payload["timeout"] = timeout
    return await _process_request(token, method_url, params=payload)


async def get_webhook_info(token, timeout=None):
    method_url = r"getWebhookInfo"
    payload = {}
    if timeout:
        payload["timeout"] = timeout
    return await _process_request(token, method_url, params=payload)


async def get_updates(
    token,
    offset=None,
    limit=None,
    timeout=None,
    allowed_updates=None,
    request_timeout=None,
):
    method_name = "getUpdates"
    params = {}
    if offset:
        params["offset"] = offset
    if limit:
        params["limit"] = limit
    if timeout:
        params["timeout"] = timeout
    if allowed_updates is not None:  # Empty lists should pass
        params["allowed_updates"] = json.dumps(allowed_updates)
    return await _process_request(
        token, method_name, params=params, request_timeout=request_timeout
    )


async def _check_result(method_name, result: aiohttp.ClientResponse):
    """
    Checks whether `result` is a valid API response.
    A result is considered invalid if:
        - The server returned an HTTP response code other than 200
        - The content of the result is invalid JSON.
        - The method call was unsuccessful (The JSON 'ok' field equals False)

    :raises ApiException: if one of the above listed cases is applicable
    :param method_name: The name of the method called
    :param result: The returned result of the method request
    :return: The result parsed to a JSON dictionary.
    """
    try:
        result_json = await result.json(encoding="utf-8")
    except:
        if result.status != 200:
            raise ApiHTTPException(method_name, result)
        else:
            raise ApiInvalidJSONException(method_name, result)
    else:
        if not result_json["ok"]:
            raise ApiTelegramException(method_name, result, result_json)

        return result_json


async def send_message(
        token, chat_id, text,
        reply_markup=None,
        parse_mode=None, disable_notification=None, timeout=None,
        entities=None, protect_content=None,
        message_thread_id=None, reply_parameters=None, link_preview_options=None, business_connection_id=None, message_effect_id=None):
    method_name = 'sendMessage'
    params = {'chat_id': str(chat_id), 'text': text}
    if link_preview_options is not None:
        params["link_preview_options"] = json.dumps(link_preview_options.to_dict())
    if reply_markup:
        params["reply_markup"] = await _convert_markup(reply_markup)
    if parse_mode:
        params["parse_mode"] = parse_mode
    if disable_notification is not None:
        params["disable_notification"] = disable_notification
    if timeout:
        params["timeout"] = timeout
    if entities:
        params["entities"] = json.dumps(types.MessageEntity.to_list_of_dicts(entities))
    if reply_parameters is not None:
        params["reply_parameters"] = json.dumps(reply_parameters.to_dict())
    if protect_content is not None:
        params["protect_content"] = protect_content
    if message_thread_id:
        params["message_thread_id"] = message_thread_id
    if business_connection_id:
        params['business_connection_id'] = business_connection_id
    if message_effect_id:
        params['message_effect_id'] = message_effect_id
    
    return await _process_request(token, method_name, params=params)


# methods


async def get_user_profile_photos(token, user_id, offset=None, limit=None):
    method_url = r"getUserProfilePhotos"
    payload = {"user_id": user_id}
    if offset:
        payload["offset"] = offset
    if limit:
        payload["limit"] = limit
    return await _process_request(token, method_url, params=payload)


async def set_message_reaction(token, chat_id, message_id, reaction=None, is_big=None):
    method_url = r"setMessageReaction"
    payload = {"chat_id": chat_id, "message_id": message_id}
    if reaction:
        payload["reaction"] = json.dumps([r.to_dict() for r in reaction])
    if is_big is not None:
        payload["is_big"] = is_big
    return await _process_request(token, method_url, params=payload)


async def get_chat(token, chat_id):
    method_url = r"getChat"
    payload = {"chat_id": chat_id}
    return await _process_request(token, method_url, params=payload)


async def leave_chat(token, chat_id):
    method_url = r"leaveChat"
    payload = {"chat_id": chat_id}
    return await _process_request(token, method_url, params=payload)


async def get_chat_administrators(token, chat_id):
    method_url = r"getChatAdministrators"
    payload = {"chat_id": chat_id}
    return await _process_request(token, method_url, params=payload)


async def get_chat_member_count(token, chat_id):
    method_url = r"getChatMemberCount"
    payload = {"chat_id": chat_id}
    return await _process_request(token, method_url, params=payload)


async def replace_sticker_in_set(token, user_id, name, old_sticker, sticker):
    method_url = r"replaceStickerInSet"
    payload = {
        "user_id": user_id,
        "name": name,
        "old_sticker": old_sticker,
        "sticker": sticker.to_json(),
    }
    return await _process_request(token, method_url, params=payload)


async def set_sticker_set_thumbnail(token, name, user_id, thumbnail, format):
    method_url = r"setStickerSetThumbnail"
    payload = {"name": name, "user_id": user_id, "format": format}
    files = {}
    if thumbnail:
        if not isinstance(thumbnail, str):
            files["thumbnail"] = thumbnail
        else:
            payload["thumbnail"] = thumbnail

    return await _process_request(
        token, method_url, params=payload, files=files or None
    )


async def set_chat_sticker_set(token, chat_id, sticker_set_name):
    method_url = r"setChatStickerSet"
    payload = {"chat_id": chat_id, "sticker_set_name": sticker_set_name}
    return await _process_request(token, method_url, params=payload)


async def delete_chat_sticker_set(token, chat_id):
    method_url = r"deleteChatStickerSet"
    payload = {"chat_id": chat_id}
    return await _process_request(token, method_url, params=payload)


async def answer_web_app_query(
    token, web_app_query_id, result: types.InlineQueryResultBase
):
    method_url = "answerWebAppQuery"
    payload = {"web_app_query_id": web_app_query_id, "result": result.to_json()}
    return await _process_request(token, method_url, params=payload, method="post")


async def get_chat_member(token, chat_id, user_id):
    method_url = r"getChatMember"
    payload = {"chat_id": chat_id, "user_id": user_id}
    return await _process_request(token, method_url, params=payload)


async def forward_message(
    token,
    chat_id,
    from_chat_id,
    message_id,
    disable_notification=None,
    timeout=None,
    protect_content=None,
    message_thread_id=None,
):
    method_url = r"forwardMessage"
    payload = {
        "chat_id": chat_id,
        "from_chat_id": from_chat_id,
        "message_id": message_id,
    }
    if disable_notification is not None:
        payload["disable_notification"] = disable_notification
    if timeout:
        payload["timeout"] = timeout
    if protect_content is not None:
        payload["protect_content"] = protect_content
    if message_thread_id:
        payload["message_thread_id"] = message_thread_id
    return await _process_request(token, method_url, params=payload)


async def copy_message(token, chat_id, from_chat_id, message_id, caption=None, parse_mode=None, caption_entities=None,
                 disable_notification=None,  
                 reply_markup=None, timeout=None, protect_content=None, message_thread_id=None, reply_parameters=None, show_caption_above_media=None):
    method_url = r'copyMessage'
    payload = {'chat_id': chat_id, 'from_chat_id': from_chat_id, 'message_id': message_id}
    if caption is not None:
        payload["caption"] = caption
    if parse_mode:
        payload["parse_mode"] = parse_mode
    if caption_entities is not None:
        payload["caption_entities"] = json.dumps(
            types.MessageEntity.to_list_of_dicts(caption_entities)
        )
    if disable_notification is not None:
        payload["disable_notification"] = disable_notification
    if reply_parameters is not None:
        payload["reply_parameters"] = json.dumps(reply_parameters.to_dict())
    if reply_markup is not None:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    if timeout:
        payload["timeout"] = timeout
    if protect_content is not None:
        payload["protect_content"] = protect_content
    if message_thread_id:
        payload['message_thread_id'] = message_thread_id
    if show_caption_above_media is not None:
        payload['show_caption_above_media'] = show_caption_above_media
    return await _process_request(token, method_url, params=payload)


async def send_dice(
        token, chat_id,
        emoji=None, disable_notification=None, 
        reply_markup=None, timeout=None,  protect_content=None,
        message_thread_id=None,reply_parameters=None, business_connection_id=None, message_effect_id=None):
    method_url = r'sendDice'
    payload = {'chat_id': chat_id}
    if emoji:
        payload["emoji"] = emoji
    if disable_notification is not None:
        payload["disable_notification"] = disable_notification
    if reply_markup:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    if timeout:
        payload["timeout"] = timeout
    if protect_content is not None:
        payload["protect_content"] = protect_content
    if message_thread_id:
        payload["message_thread_id"] = message_thread_id
    if reply_parameters is not None:
        payload["reply_parameters"] = json.dumps(reply_parameters.to_dict())
    if business_connection_id:
        payload['business_connection_id'] = business_connection_id
    if message_effect_id:
        payload['message_effect_id'] = message_effect_id
    return await _process_request(token, method_url, params=payload)


async def send_photo(
        token, chat_id, photo,
        caption=None,  reply_markup=None,
        parse_mode=None, disable_notification=None, timeout=None,
        caption_entities=None,  protect_content=None,
        message_thread_id=None, has_spoiler=None,reply_parameters=None,
        business_connection_id=None, message_effect_id=None, show_caption_above_media=None):
    method_url = r'sendPhoto'
    payload = {'chat_id': chat_id}
    files = None
    if util.is_string(photo):
        payload["photo"] = photo
    elif util.is_pil_image(photo):
        files = {"photo": util.pil_image_to_file(photo)}
    else:
        files = {"photo": photo}
    if caption:
        payload["caption"] = caption
    if reply_markup:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    if parse_mode:
        payload["parse_mode"] = parse_mode
    if disable_notification is not None:
        payload["disable_notification"] = disable_notification
    if timeout:
        payload["timeout"] = timeout
    if caption_entities:
        payload["caption_entities"] = json.dumps(
            types.MessageEntity.to_list_of_dicts(caption_entities)
        )
    if reply_parameters is not None:
        payload["reply_parameters"] = json.dumps(reply_parameters.to_dict())
    if protect_content is not None:
        payload["protect_content"] = protect_content
    if message_thread_id:
        payload["message_thread_id"] = message_thread_id
    if has_spoiler is not None:
        payload["has_spoiler"] = has_spoiler
    if business_connection_id:
        payload['business_connection_id'] = business_connection_id
    if message_effect_id:
        payload['message_effect_id'] = message_effect_id
    if show_caption_above_media is not None:
        payload['show_caption_above_media'] = show_caption_above_media
    return await _process_request(token, method_url, params=payload, files=files, method='post')


async def send_media_group(
        token, chat_id, media,
        disable_notification=None, 
        timeout=None,  protect_content=None, message_thread_id=None,reply_parameters=None, business_connection_id=None, message_effect_id=None):
    method_url = r'sendMediaGroup'
    media_json, files = await convert_input_media_array(media)
    payload = {"chat_id": chat_id, "media": media_json}
    if disable_notification is not None:
        payload["disable_notification"] = disable_notification
    if timeout:
        payload["timeout"] = timeout
    if protect_content is not None:
        payload["protect_content"] = protect_content
    if message_thread_id:
        payload["message_thread_id"] = message_thread_id
    if reply_parameters is not None:
        payload["reply_parameters"] = json.dumps(reply_parameters.to_dict())
    if business_connection_id:
        payload['business_connection_id'] = business_connection_id
    if message_effect_id:
        payload['message_effect_id'] = message_effect_id
    return await _process_request(
        token,
        method_url,
        params=payload,
        method="post" if files else "get",
        files=files if files else None,
    )


async def send_location(
        token, chat_id, latitude, longitude,
        live_period=None,  
        reply_markup=None, disable_notification=None, 
        timeout=None, horizontal_accuracy=None, heading=None,
        proximity_alert_radius=None,  protect_content=None, message_thread_id=None,reply_parameters=None, business_connection_id=None,
        message_effect_id=None):
    method_url = r'sendLocation'
    payload = {'chat_id': chat_id, 'latitude': latitude, 'longitude': longitude}
    if live_period:
        payload["live_period"] = live_period
    if horizontal_accuracy:
        payload["horizontal_accuracy"] = horizontal_accuracy
    if heading:
        payload["heading"] = heading
    if proximity_alert_radius:
        payload["proximity_alert_radius"] = proximity_alert_radius
    if reply_parameters is not None:
        payload["reply_parameters"] = json.dumps(reply_parameters.to_dict())
    if reply_markup:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    if disable_notification is not None:
        payload["disable_notification"] = disable_notification
    if timeout:
        payload["timeout"] = timeout
    if protect_content is not None:
        payload["protect_content"] = protect_content
    if message_thread_id:
        payload["message_thread_id"] = message_thread_id
    if business_connection_id:
        payload['business_connection_id'] = business_connection_id
    if message_effect_id:
        payload['message_effect_id'] = message_effect_id
    return await _process_request(token, method_url, params=payload)


async def edit_message_live_location(
        token, latitude, longitude, chat_id=None, message_id=None, inline_message_id=None, reply_markup=None,
        timeout=None, horizontal_accuracy=None, heading=None, proximity_alert_radius=None, live_period=None):
    method_url = r'editMessageLiveLocation'
    payload = {'latitude': latitude, 'longitude': longitude}
    if chat_id:
        payload["chat_id"] = chat_id
    if message_id:
        payload["message_id"] = message_id
    if horizontal_accuracy:
        payload["horizontal_accuracy"] = horizontal_accuracy
    if heading:
        payload["heading"] = heading
    if proximity_alert_radius:
        payload['proximity_alert_radius'] = proximity_alert_radius
    if live_period:
        payload['live_period'] = live_period
    if inline_message_id:
        payload["inline_message_id"] = inline_message_id
    if reply_markup:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    if timeout:
        payload["timeout"] = timeout
    return await _process_request(token, method_url, params=payload)


async def stop_message_live_location(
    token,
    chat_id=None,
    message_id=None,
    inline_message_id=None,
    reply_markup=None,
    timeout=None,
):
    method_url = r"stopMessageLiveLocation"
    payload = {}
    if chat_id:
        payload["chat_id"] = chat_id
    if message_id:
        payload["message_id"] = message_id
    if inline_message_id:
        payload["inline_message_id"] = inline_message_id
    if reply_markup:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    if timeout:
        payload["timeout"] = timeout
    return await _process_request(token, method_url, params=payload)


async def send_venue(
        token, chat_id, latitude, longitude, title, address,
        foursquare_id=None, foursquare_type=None, disable_notification=None,
         reply_markup=None, timeout=None,
         google_place_id=None,
        google_place_type=None, protect_content=None, message_thread_id=None,reply_parameters=None, business_connection_id=None,
        message_effect_id=None):
    method_url = r'sendVenue'
    payload = {'chat_id': chat_id, 'latitude': latitude, 'longitude': longitude, 'title': title, 'address': address}
    if foursquare_id:
        payload["foursquare_id"] = foursquare_id
    if foursquare_type:
        payload["foursquare_type"] = foursquare_type
    if disable_notification is not None:
        payload["disable_notification"] = disable_notification
    if reply_parameters is not None:
        payload["reply_parameters"] = json.dumps(reply_parameters.to_dict())
    if reply_markup:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    if timeout:
        payload["timeout"] = timeout
    if google_place_id:
        payload["google_place_id"] = google_place_id
    if google_place_type:
        payload["google_place_type"] = google_place_type
    if protect_content is not None:
        payload["protect_content"] = protect_content
    if message_thread_id:
        payload["message_thread_id"] = message_thread_id
    if business_connection_id:
        payload['business_connection_id'] = business_connection_id
    if message_effect_id:
        payload['message_effect_id'] = message_effect_id
    return await _process_request(token, method_url, params=payload)


async def send_contact(
        token, chat_id, phone_number, first_name, last_name=None, vcard=None,
        disable_notification=None,  reply_markup=None, timeout=None,
         protect_content=None, message_thread_id=None,reply_parameters=None, business_connection_id=None, message_effect_id=None):
    method_url = r'sendContact'
    payload = {'chat_id': chat_id, 'phone_number': phone_number, 'first_name': first_name}
    if last_name:
        payload["last_name"] = last_name
    if vcard:
        payload["vcard"] = vcard
    if disable_notification is not None:
        payload["disable_notification"] = disable_notification
    if reply_parameters is not None:
        payload["reply_parameters"] = json.dumps(reply_parameters.to_dict())
    if reply_markup:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    if timeout:
        payload["timeout"] = timeout
    if protect_content is not None:
        payload["protect_content"] = protect_content
    if message_thread_id:
        payload["message_thread_id"] = message_thread_id
    if business_connection_id:
        payload['business_connection_id'] = business_connection_id
    if message_effect_id:
        payload['message_effect_id'] = message_effect_id
    return await _process_request(token, method_url, params=payload)


async def send_chat_action(
    token,
    chat_id,
    action,
    timeout=None,
    message_thread_id=None,
    business_connection_id=None,
):
    method_url = r"sendChatAction"
    payload = {"chat_id": chat_id, "action": action}
    if timeout:
        payload["timeout"] = timeout
    if message_thread_id:
        payload["message_thread_id"] = message_thread_id
    if business_connection_id:
        payload["business_connection_id"] = business_connection_id
    return await _process_request(token, method_url, params=payload)


async def send_video(token, chat_id, data, duration=None, caption=None,  reply_markup=None,
                     parse_mode=None, supports_streaming=None, disable_notification=None, timeout=None,
                     thumbnail=None, width=None, height=None, caption_entities=None, 
                     protect_content=None, message_thread_id=None, has_spoiler=None,reply_parameters=None, business_connection_id=None,
                     message_effect_id=None, show_caption_above_media=None):
    method_url = r'sendVideo'
    payload = {'chat_id': chat_id}
    files = None
    if not util.is_string(data):
        files = {"video": data}
    else:
        payload["video"] = data
    if duration:
        payload["duration"] = duration
    if caption:
        payload["caption"] = caption
    if reply_parameters is not None:
        payload["reply_parameters"] = json.dumps(reply_parameters.to_dict())
    if reply_markup:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    if parse_mode:
        payload["parse_mode"] = parse_mode
    if supports_streaming is not None:
        payload["supports_streaming"] = supports_streaming
    if disable_notification is not None:
        payload["disable_notification"] = disable_notification
    if timeout:
        payload["timeout"] = timeout
    if thumbnail:
        if not util.is_string(thumbnail):
            if files:
                files["thumbnail"] = thumbnail
            else:
                files = {"thumbnail": thumbnail}
        else:
            payload["thumbnail"] = thumbnail
    if width:
        payload["width"] = width
    if height:
        payload["height"] = height
    if caption_entities:
        payload["caption_entities"] = json.dumps(
            types.MessageEntity.to_list_of_dicts(caption_entities)
        )
    if protect_content is not None:
        payload["protect_content"] = protect_content
    if message_thread_id:
        payload["message_thread_id"] = message_thread_id
    if has_spoiler is not None:
        payload["has_spoiler"] = has_spoiler
    if business_connection_id:
        payload['business_connection_id'] = business_connection_id
    if message_effect_id:
        payload['message_effect_id'] = message_effect_id
    if show_caption_above_media is not None:
        payload['show_caption_above_media'] = show_caption_above_media
    return await _process_request(token, method_url, params=payload, files=files, method='post')


async def send_animation(
        token, chat_id, data, duration=None, caption=None,  reply_markup=None,
        parse_mode=None, disable_notification=None, timeout=None, thumbnail=None, caption_entities=None,
         width=None, height=None, protect_content=None, message_thread_id=None,
        has_spoiler=None,reply_parameters=None, business_connection_id=None, message_effect_id=None, show_caption_above_media=None):
    method_url = r'sendAnimation'
    payload = {'chat_id': chat_id}
    files = None
    if not util.is_string(data):
        files = {"animation": data}
    else:
        payload["animation"] = data
    if duration:
        payload["duration"] = duration
    if caption:
        payload["caption"] = caption
    if reply_parameters is not None:
        payload["reply_parameters"] = json.dumps(reply_parameters.to_dict())
    if reply_markup:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    if parse_mode:
        payload["parse_mode"] = parse_mode
    if disable_notification is not None:
        payload["disable_notification"] = disable_notification
    if timeout:
        payload["timeout"] = timeout
    if thumbnail:
        if not util.is_string(thumbnail):
            if files:
                files["thumbnail"] = thumbnail
            else:
                files = {"thumbnail": thumbnail}
        else:
            payload["thumbnail"] = thumbnail
    if caption_entities:
        payload["caption_entities"] = json.dumps(
            types.MessageEntity.to_list_of_dicts(caption_entities)
        )
    if width:
        payload["width"] = width
    if height:
        payload["height"] = height
    if protect_content is not None:
        payload["protect_content"] = protect_content
    if message_thread_id:
        payload["message_thread_id"] = message_thread_id
    if has_spoiler is not None:
        payload["has_spoiler"] = has_spoiler
    if business_connection_id:
        payload['business_connection_id'] = business_connection_id
    if message_effect_id:
        payload['message_effect_id'] = message_effect_id
    if show_caption_above_media is not None:
        payload['show_caption_above_media'] = show_caption_above_media
    return await _process_request(token, method_url, params=payload, files=files, method='post')


async def send_voice(token, chat_id, voice, caption=None, duration=None,  reply_markup=None,
               parse_mode=None, disable_notification=None, timeout=None, caption_entities=None,
                protect_content=None, message_thread_id=None,reply_parameters=None,business_connection_id=None, message_effect_id=None):
    method_url = r'sendVoice'
    payload = {'chat_id': chat_id}
    files = None
    if not util.is_string(voice):
        files = {"voice": voice}
    else:
        payload["voice"] = voice
    if caption:
        payload["caption"] = caption
    if duration:
        payload["duration"] = duration
    if reply_parameters is not None:
        payload["reply_parameters"] = json.dumps(reply_parameters.to_dict())
    if reply_markup:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    if parse_mode:
        payload["parse_mode"] = parse_mode
    if disable_notification is not None:
        payload["disable_notification"] = disable_notification
    if timeout:
        payload["timeout"] = timeout
    if caption_entities:
        payload["caption_entities"] = json.dumps(
            types.MessageEntity.to_list_of_dicts(caption_entities)
        )
    if protect_content is not None:
        payload["protect_content"] = protect_content
    if message_thread_id:
        payload["message_thread_id"] = message_thread_id
    if business_connection_id:
        payload['business_connection_id'] = business_connection_id
    if message_effect_id:
        payload['message_effect_id'] = message_effect_id
    return await _process_request(token, method_url, params=payload, files=files, method='post')


async def send_video_note(token, chat_id, data, duration=None, length=None,  reply_markup=None,
                          disable_notification=None, timeout=None, thumbnail=None,  protect_content=None,
                          message_thread_id=None,reply_parameters=None, business_connection_id=None, message_effect_id=None):
    method_url = r'sendVideoNote'
    payload = {'chat_id': chat_id}
    files = None
    if not util.is_string(data):
        files = {"video_note": data}
    else:
        payload["video_note"] = data
    if duration:
        payload["duration"] = duration
    if length and (str(length).isdigit() and int(length) <= 639):
        payload["length"] = length
    else:
        payload["length"] = 639  # seems like it is MAX length size
    if reply_parameters is not None:
        payload["reply_parameters"] = json.dumps(reply_parameters.to_dict())
    if reply_markup:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    if disable_notification is not None:
        payload["disable_notification"] = disable_notification
    if timeout:
        payload["timeout"] = timeout
    if thumbnail:
        if not util.is_string(thumbnail):
            if files:
                files["thumbnail"] = thumbnail
            else:
                files = {"thumbnail": thumbnail}
        else:
            payload["thumbnail"] = thumbnail
    if protect_content is not None:
        payload["protect_content"] = protect_content
    if message_thread_id:
        payload["message_thread_id"] = message_thread_id
    if business_connection_id:
        payload['business_connection_id'] = business_connection_id
    if message_effect_id:
        payload['message_effect_id'] = message_effect_id
    return await _process_request(token, method_url, params=payload, files=files, method='post')


async def send_audio(token, chat_id, audio, caption=None, duration=None, performer=None, title=None, 
                     reply_markup=None, parse_mode=None, disable_notification=None, timeout=None, thumbnail=None,
                     caption_entities=None,  protect_content=None, message_thread_id=None,reply_parameters=None, business_connection_id=None,
                     message_effect_id=None):
    method_url = r'sendAudio'
    payload = {'chat_id': chat_id}
    files = None
    if not util.is_string(audio):
        files = {"audio": audio}
    else:
        payload["audio"] = audio
    if caption:
        payload["caption"] = caption
    if duration:
        payload["duration"] = duration
    if performer:
        payload["performer"] = performer
    if title:
        payload["title"] = title
    if reply_parameters is not None:
        payload["reply_parameters"] = json.dumps(reply_parameters.to_dict())
    if reply_markup:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    if parse_mode:
        payload["parse_mode"] = parse_mode
    if disable_notification is not None:
        payload["disable_notification"] = disable_notification
    if timeout:
        payload["timeout"] = timeout
    if thumbnail:
        if not util.is_string(thumbnail):
            if files:
                files["thumbnail"] = thumbnail
            else:
                files = {"thumbnail": thumbnail}
        else:
            payload["thumbnail"] = thumbnail
    if caption_entities:
        payload["caption_entities"] = json.dumps(
            types.MessageEntity.to_list_of_dicts(caption_entities)
        )
    if protect_content is not None:
        payload["protect_content"] = protect_content
    if message_thread_id:
        payload["message_thread_id"] = message_thread_id
    if business_connection_id:
        payload['business_connection_id'] = business_connection_id
    if message_effect_id:
        payload['message_effect_id'] = message_effect_id
    return await _process_request(token, method_url, params=payload, files=files, method='post')


async def send_data(token, chat_id, data, data_type,  reply_markup=None, parse_mode=None,
                    disable_notification=None, timeout=None, caption=None, thumbnail=None, caption_entities=None,
                     disable_content_type_detection=None, visible_file_name=None, protect_content=None,
                    message_thread_id=None, emoji=None,reply_parameters=None, business_connection_id=None, message_effect_id=None):
    method_url = await get_method_by_type(data_type)
    payload = {"chat_id": chat_id}
    files = None
    if not util.is_string(data):
        file_data = data
        if visible_file_name:
            file_data = (visible_file_name, data)
        files = {data_type: file_data}
    else:
        payload[data_type] = data
    if reply_parameters is not None:
        payload["reply_parameters"] = json.dumps(reply_parameters.to_dict())
    if reply_markup:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    if parse_mode and data_type == "document":
        payload["parse_mode"] = parse_mode
    if disable_notification is not None:
        payload["disable_notification"] = disable_notification
    if timeout:
        payload["timeout"] = timeout
    if caption:
        payload["caption"] = caption
    if thumbnail:
        if not util.is_string(thumbnail):
            if files:
                files["thumbnail"] = thumbnail
            else:
                files = {"thumbnail": thumbnail}
        else:
            payload["thumbnail"] = thumbnail
    if caption_entities:
        payload["caption_entities"] = json.dumps(
            types.MessageEntity.to_list_of_dicts(caption_entities)
        )
    if protect_content is not None:
        payload["protect_content"] = protect_content
    if method_url == "sendDocument" and disable_content_type_detection is not None:
        payload["disable_content_type_detection"] = disable_content_type_detection
    if message_thread_id:
        payload["message_thread_id"] = message_thread_id
    if emoji:
        payload["emoji"] = emoji
    if business_connection_id:
        payload['business_connection_id'] = business_connection_id
    if message_effect_id:
        payload['message_effect_id'] = message_effect_id
    return await _process_request(token, method_url, params=payload, files=files, method='post')


async def get_method_by_type(data_type):
    if data_type == "document":
        return r"sendDocument"
    if data_type == "sticker":
        return r"sendSticker"


async def ban_chat_member(
    token, chat_id, user_id, until_date=None, revoke_messages=None
):
    method_url = "banChatMember"
    payload = {"chat_id": chat_id, "user_id": user_id}
    if isinstance(until_date, datetime):
        payload["until_date"] = until_date.timestamp()
    else:
        payload["until_date"] = until_date
    if revoke_messages is not None:
        payload["revoke_messages"] = revoke_messages
    return await _process_request(token, method_url, params=payload, method="post")


async def unban_chat_member(token, chat_id, user_id, only_if_banned):
    method_url = "unbanChatMember"
    payload = {"chat_id": chat_id, "user_id": user_id}
    if only_if_banned is not None:  # None / True / False
        payload["only_if_banned"] = only_if_banned
    return await _process_request(token, method_url, params=payload, method="post")


async def restrict_chat_member(
    token,
    chat_id,
    user_id,
    permissions,
    until_date=None,
    use_independent_chat_permissions=None,
):
    method_url = "restrictChatMember"
    payload = {
        "chat_id": chat_id,
        "user_id": user_id,
        "permissions": permissions.to_json(),
    }

    if use_independent_chat_permissions is not None:
        payload["use_independent_chat_permissions"] = use_independent_chat_permissions
    if until_date is not None:
        if isinstance(until_date, datetime):
            payload["until_date"] = until_date.timestamp()
        else:
            payload["until_date"] = until_date
    return await _process_request(token, method_url, params=payload, method="post")


async def promote_chat_member(
    token,
    chat_id,
    user_id,
    can_change_info=None,
    can_post_messages=None,
    can_edit_messages=None,
    can_delete_messages=None,
    can_invite_users=None,
    can_restrict_members=None,
    can_pin_messages=None,
    can_promote_members=None,
    is_anonymous=None,
    can_manage_chat=None,
    can_manage_video_chats=None,
    can_manage_topics=None,
    can_post_stories=None,
    can_edit_stories=None,
    can_delete_stories=None,
):
    method_url = "promoteChatMember"
    payload = {"chat_id": chat_id, "user_id": user_id}
    if can_change_info is not None:
        payload["can_change_info"] = can_change_info
    if can_post_messages is not None:
        payload["can_post_messages"] = can_post_messages
    if can_edit_messages is not None:
        payload["can_edit_messages"] = can_edit_messages
    if can_delete_messages is not None:
        payload["can_delete_messages"] = can_delete_messages
    if can_invite_users is not None:
        payload["can_invite_users"] = can_invite_users
    if can_restrict_members is not None:
        payload["can_restrict_members"] = can_restrict_members
    if can_pin_messages is not None:
        payload["can_pin_messages"] = can_pin_messages
    if can_promote_members is not None:
        payload["can_promote_members"] = can_promote_members
    if is_anonymous is not None:
        payload["is_anonymous"] = is_anonymous
    if can_manage_chat is not None:
        payload["can_manage_chat"] = can_manage_chat
    if can_manage_video_chats is not None:
        payload["can_manage_video_chats"] = can_manage_video_chats
    if can_manage_topics is not None:
        payload["can_manage_topics"] = can_manage_topics
    if can_post_stories is not None:
        payload["can_post_stories"] = can_post_stories
    if can_edit_stories is not None:
        payload["can_edit_stories"] = can_edit_stories
    if can_delete_stories is not None:
        payload["can_delete_stories"] = can_delete_stories
    return await _process_request(token, method_url, params=payload, method="post")


async def set_chat_administrator_custom_title(token, chat_id, user_id, custom_title):
    method_url = "setChatAdministratorCustomTitle"
    payload = {"chat_id": chat_id, "user_id": user_id, "custom_title": custom_title}
    return await _process_request(token, method_url, params=payload, method="post")


async def ban_chat_sender_chat(token, chat_id, sender_chat_id):
    method_url = "banChatSenderChat"
    payload = {"chat_id": chat_id, "sender_chat_id": sender_chat_id}
    return await _process_request(token, method_url, params=payload, method="post")


async def unban_chat_sender_chat(token, chat_id, sender_chat_id):
    method_url = "unbanChatSenderChat"
    payload = {"chat_id": chat_id, "sender_chat_id": sender_chat_id}
    return await _process_request(token, method_url, params=payload, method="post")


async def set_chat_permissions(
    token, chat_id, permissions, use_independent_chat_permissions=None
):
    method_url = "setChatPermissions"
    payload = {"chat_id": chat_id, "permissions": permissions.to_json()}
    if use_independent_chat_permissions is not None:
        payload["use_independent_chat_permissions"] = use_independent_chat_permissions
    return await _process_request(token, method_url, params=payload, method="post")


async def create_chat_invite_link(
    token, chat_id, name, expire_date, member_limit, creates_join_request
):
    method_url = "createChatInviteLink"
    payload = {"chat_id": chat_id}

    if expire_date is not None:
        if isinstance(expire_date, datetime):
            payload["expire_date"] = expire_date.timestamp()
        else:
            payload["expire_date"] = expire_date
    if member_limit:
        payload["member_limit"] = member_limit
    if creates_join_request is not None:
        payload["creates_join_request"] = creates_join_request
    if name:
        payload["name"] = name

    return await _process_request(token, method_url, params=payload, method="post")


async def edit_chat_invite_link(
    token, chat_id, invite_link, name, expire_date, member_limit, creates_join_request
):
    method_url = "editChatInviteLink"
    payload = {"chat_id": chat_id, "invite_link": invite_link}

    if expire_date is not None:
        if isinstance(expire_date, datetime):
            payload["expire_date"] = expire_date.timestamp()
        else:
            payload["expire_date"] = expire_date

    if member_limit is not None:
        payload["member_limit"] = member_limit
    if name:
        payload["name"] = name
    if creates_join_request is not None:
        payload["creates_join_request"] = creates_join_request

    return await _process_request(token, method_url, params=payload, method="post")


async def revoke_chat_invite_link(token, chat_id, invite_link):
    method_url = "revokeChatInviteLink"
    payload = {"chat_id": chat_id, "invite_link": invite_link}
    return await _process_request(token, method_url, params=payload, method="post")


async def export_chat_invite_link(token, chat_id):
    method_url = "exportChatInviteLink"
    payload = {"chat_id": chat_id}
    return await _process_request(token, method_url, params=payload, method="post")


async def approve_chat_join_request(token, chat_id, user_id):
    method_url = "approveChatJoinRequest"
    payload = {"chat_id": chat_id, "user_id": user_id}
    return await _process_request(token, method_url, params=payload, method="post")


async def decline_chat_join_request(token, chat_id, user_id):
    method_url = "declineChatJoinRequest"
    payload = {"chat_id": chat_id, "user_id": user_id}
    return await _process_request(token, method_url, params=payload, method="post")


async def set_chat_photo(token, chat_id, photo):
    method_url = "setChatPhoto"
    payload = {"chat_id": chat_id}
    files = None
    if util.is_string(photo):
        payload["photo"] = photo
    elif util.is_pil_image(photo):
        files = {"photo": util.pil_image_to_file(photo)}
    else:
        files = {"photo": photo}
    return await _process_request(
        token, method_url, params=payload, files=files, method="post"
    )


async def delete_chat_photo(token, chat_id):
    method_url = "deleteChatPhoto"
    payload = {"chat_id": chat_id}
    return await _process_request(token, method_url, params=payload, method="post")


async def set_chat_title(token, chat_id, title):
    method_url = "setChatTitle"
    payload = {"chat_id": chat_id, "title": title}
    return await _process_request(token, method_url, params=payload, method="post")


async def set_my_description(token, description=None, language_code=None):
    method_url = r"setMyDescription"
    payload = {}
    if description is not None:
        payload["description"] = description
    if language_code is not None:
        payload["language_code"] = language_code
    return await _process_request(token, method_url, params=payload, method="post")


async def get_my_description(token, language_code=None):
    method_url = r"getMyDescription"
    payload = {}
    if language_code:
        payload["language_code"] = language_code
    return await _process_request(token, method_url, params=payload)


async def set_my_short_description(token, short_description=None, language_code=None):
    method_url = r"setMyShortDescription"
    payload = {}
    if short_description is not None:
        payload["short_description"] = short_description
    if language_code is not None:
        payload["language_code"] = language_code
    return await _process_request(token, method_url, params=payload, method="post")


async def get_my_short_description(token, language_code=None):
    method_url = r"getMyShortDescription"
    payload = {}
    if language_code:
        payload["language_code"] = language_code
    return await _process_request(token, method_url, params=payload)


async def get_my_commands(token, scope=None, language_code=None):
    method_url = r"getMyCommands"
    payload = {}
    if scope:
        payload["scope"] = scope.to_json()
    if language_code:
        payload["language_code"] = language_code
    return await _process_request(token, method_url, params=payload)


async def set_my_name(token, name=None, language_code=None):
    method_url = r"setMyName"
    payload = {}
    if name is not None:
        payload["name"] = name
    if language_code is not None:
        payload["language_code"] = language_code
    return await _process_request(token, method_url, params=payload, method="post")


async def get_my_name(token, language_code=None):
    method_url = r"getMyName"
    payload = {}
    if language_code is not None:
        payload["language_code"] = language_code
    return await _process_request(token, method_url, params=payload)


async def set_chat_menu_button(token, chat_id=None, menu_button=None):
    method_url = r"setChatMenuButton"
    payload = {}
    if chat_id:
        payload["chat_id"] = chat_id
    if menu_button:
        payload["menu_button"] = menu_button.to_json()

    return await _process_request(token, method_url, params=payload, method="post")


async def get_chat_menu_button(token, chat_id=None):
    method_url = r"getChatMenuButton"
    payload = {}
    if chat_id:
        payload["chat_id"] = chat_id

    return await _process_request(token, method_url, params=payload, method="post")


async def set_my_default_administrator_rights(token, rights=None, for_channels=None):
    method_url = r"setMyDefaultAdministratorRights"
    payload = {}
    if rights:
        payload["rights"] = rights.to_json()
    if for_channels is not None:
        payload["for_channels"] = for_channels

    return await _process_request(token, method_url, params=payload, method="post")


async def get_my_default_administrator_rights(token, for_channels=None):
    method_url = r"getMyDefaultAdministratorRights"
    payload = {}
    if for_channels:
        payload["for_channels"] = for_channels

    return await _process_request(token, method_url, params=payload, method="post")


async def set_my_commands(token, commands, scope=None, language_code=None):
    method_url = r"setMyCommands"
    payload = {"commands": await _convert_list_json_serializable(commands)}
    if scope:
        payload["scope"] = scope.to_json()
    if language_code:
        payload["language_code"] = language_code
    return await _process_request(token, method_url, params=payload, method="post")


async def get_business_connection(token, business_connection_id):
    method_url = "getBusinessConnection"
    payload = {"business_connection_id": business_connection_id}
    return await _process_request(token, method_url, params=payload, method="post")


async def delete_my_commands(token, scope=None, language_code=None):
    method_url = r"deleteMyCommands"
    payload = {}
    if scope:
        payload["scope"] = scope.to_json()
    if language_code:
        payload["language_code"] = language_code
    return await _process_request(token, method_url, params=payload, method="post")


async def set_chat_description(token, chat_id, description):
    method_url = "setChatDescription"
    payload = {"chat_id": chat_id}
    if description is not None:  # Allow empty strings
        payload["description"] = description
    return await _process_request(token, method_url, params=payload, method="post")


async def pin_chat_message(token, chat_id, message_id, disable_notification=None):
    method_url = "pinChatMessage"
    payload = {"chat_id": chat_id, "message_id": message_id}
    if disable_notification is not None:
        payload["disable_notification"] = disable_notification
    return await _process_request(token, method_url, params=payload, method="post")


async def unpin_chat_message(token, chat_id, message_id):
    method_url = "unpinChatMessage"
    payload = {"chat_id": chat_id}
    if message_id:
        payload["message_id"] = message_id
    return await _process_request(token, method_url, params=payload, method="post")


async def unpin_all_chat_messages(token, chat_id):
    method_url = "unpinAllChatMessages"
    payload = {"chat_id": chat_id}
    return await _process_request(token, method_url, params=payload, method="post")


# Updating messages


async def edit_message_text(
    token,
    text,
    chat_id=None,
    message_id=None,
    inline_message_id=None,
    parse_mode=None,
    entities=None,
    reply_markup=None,
    link_preview_options=None,
):
    method_url = r"editMessageText"
    payload = {"text": text}
    if chat_id:
        payload["chat_id"] = chat_id
    if message_id:
        payload["message_id"] = message_id
    if inline_message_id:
        payload["inline_message_id"] = inline_message_id
    if parse_mode:
        payload["parse_mode"] = parse_mode
    if entities:
        payload["entities"] = json.dumps(types.MessageEntity.to_list_of_dicts(entities))
    if reply_markup:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    if link_preview_options is not None:
        payload['link_preview_options'] = link_preview_options.to_json()
    return await _process_request(token, method_url, params=payload, method='post')


async def edit_message_caption(token, caption, chat_id=None, message_id=None, inline_message_id=None,
                         parse_mode=None, caption_entities=None,reply_markup=None, show_caption_above_media=None):
    method_url = r'editMessageCaption'
    payload = {'caption': caption}
    if chat_id:
        payload["chat_id"] = chat_id
    if message_id:
        payload["message_id"] = message_id
    if inline_message_id:
        payload["inline_message_id"] = inline_message_id
    if parse_mode:
        payload["parse_mode"] = parse_mode
    if caption_entities:
        payload["caption_entities"] = json.dumps(
            types.MessageEntity.to_list_of_dicts(caption_entities)
        )
    if reply_markup:
        payload['reply_markup'] = await _convert_markup(reply_markup)
    if show_caption_above_media is not None:
        payload['show_caption_above_media'] = show_caption_above_media
    return await _process_request(token, method_url, params=payload, method='post')


async def edit_message_media(token, media, chat_id=None, message_id=None, inline_message_id=None, reply_markup=None):
    method_url = r'editMessageMedia'
    media_json, file = await convert_input_media(media)
    payload = {"media": media_json}
    if chat_id:
        payload["chat_id"] = chat_id
    if message_id:
        payload["message_id"] = message_id
    if inline_message_id:
        payload["inline_message_id"] = inline_message_id
    if reply_markup:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    return await _process_request(
        token, method_url, params=payload, files=file, method="post" if file else "get"
    )


async def edit_message_reply_markup(
    token, chat_id=None, message_id=None, inline_message_id=None, reply_markup=None
):
    method_url = r"editMessageReplyMarkup"
    payload = {}
    if chat_id:
        payload["chat_id"] = chat_id
    if message_id:
        payload["message_id"] = message_id
    if inline_message_id:
        payload["inline_message_id"] = inline_message_id
    if reply_markup:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    return await _process_request(token, method_url, params=payload, method="post")


async def delete_message(token, chat_id, message_id, timeout=None):
    method_url = r"deleteMessage"
    payload = {"chat_id": chat_id, "message_id": message_id}
    if timeout:
        payload["timeout"] = timeout
    return await _process_request(token, method_url, params=payload, method="post")


# Game


async def send_game(
        token, chat_id, game_short_name,
        disable_notification=None,  reply_markup=None, timeout=None,
         protect_content=None, message_thread_id=None,reply_parameters=None, business_connection_id=None, message_effect_id=None):
    method_url = r'sendGame'
    payload = {'chat_id': chat_id, 'game_short_name': game_short_name}
    if disable_notification is not None:
        payload["disable_notification"] = disable_notification
    if reply_parameters is not None:
        payload["reply_parameters"] = json.dumps(reply_parameters.to_dict())
    if reply_markup:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    if timeout:
        payload["timeout"] = timeout
    if protect_content is not None:
        payload["protect_content"] = protect_content
    if message_thread_id:
        payload["message_thread_id"] = message_thread_id
    if business_connection_id:
        payload['business_connection_id'] = business_connection_id
    if message_effect_id:
        payload['message_effect_id'] = message_effect_id
    return await _process_request(token, method_url, params=payload)


# https://core.telegram.org/bots/api#setgamescore
async def set_game_score(
    token,
    user_id,
    score,
    force=None,
    disable_edit_message=None,
    chat_id=None,
    message_id=None,
    inline_message_id=None,
):
    """
    Use this method to set the score of the specified user in a game. On success, if the message was sent by the bot, returns the edited Message, otherwise returns True. Returns an error, if the new score is not greater than the user's current score in the chat.
    :param token: Bot's token (you don't need to fill this)
    :param user_id: User identifier
    :param score: New score, must be non-negative
    :param force: (Optional) Pass True, if the high score is allowed to decrease. This can be useful when fixing mistakes or banning cheaters
    :param disable_edit_message: (Optional) Pass True, if the game message should not be automatically edited to include the current scoreboard
    :param chat_id: (Optional, required if inline_message_id is not specified) Unique identifier for the target chat (or username of the target channel in the format @channelusername)
    :param message_id: (Optional, required if inline_message_id is not specified) Unique identifier of the sent message
    :param inline_message_id: (Optional, required if chat_id and message_id are not specified) Identifier of the inline message
    :return:
    """
    method_url = r"setGameScore"
    payload = {"user_id": user_id, "score": score}
    if force is not None:
        payload["force"] = force
    if chat_id:
        payload["chat_id"] = chat_id
    if message_id:
        payload["message_id"] = message_id
    if inline_message_id:
        payload["inline_message_id"] = inline_message_id
    if disable_edit_message is not None:
        payload["disable_edit_message"] = disable_edit_message
    return await _process_request(token, method_url, params=payload)


# https://core.telegram.org/bots/api#getgamehighscores
async def get_game_high_scores(
    token, user_id, chat_id=None, message_id=None, inline_message_id=None
):
    """
    Use this method to get data for high score tables. Will return the score of the specified user and several of his neighbors in a game. On success, returns an Array of GameHighScore objects.
    This method will currently return scores for the target user, plus two of his closest neighbors on each side. Will also return the top three users if the user and his neighbors are not among them. Please note that this behavior is subject to change.
    :param token: Bot's token (you don't need to fill this)
    :param user_id: Target user id
    :param chat_id: (Optional, required if inline_message_id is not specified) Unique identifier for the target chat (or username of the target channel in the format @channelusername)
    :param message_id: (Optional, required if inline_message_id is not specified) Unique identifier of the sent message
    :param inline_message_id: (Optional, required if chat_id and message_id are not specified) Identifier of the inline message
    :return:
    """
    method_url = r"getGameHighScores"
    payload = {"user_id": user_id}
    if chat_id:
        payload["chat_id"] = chat_id
    if message_id:
        payload["message_id"] = message_id
    if inline_message_id:
        payload["inline_message_id"] = inline_message_id
    return await _process_request(token, method_url, params=payload)


# Payments (https://core.telegram.org/bots/api#payments)


async def send_invoice(
        token, chat_id, title, description, invoice_payload, provider_token, currency, prices,
        start_parameter = None, photo_url=None, photo_size=None, photo_width=None, photo_height=None,
        need_name=None, need_phone_number=None, need_email=None, need_shipping_address=None,
        send_phone_number_to_provider = None, send_email_to_provider = None, is_flexible=None,
        disable_notification=None,  reply_markup=None, provider_data=None,
        timeout=None,  max_tip_amount=None, suggested_tip_amounts=None,
        protect_content=None, message_thread_id=None, reply_parameters=None, message_effect_id=None):
    """
    Use this method to send invoices. On success, the sent Message is returned.
    :param token: Bot's token (you don't need to fill this)
    :param chat_id: Unique identifier for the target private chat
    :param title: Product name
    :param description: Product description
    :param invoice_payload: Bot-defined invoice payload, 1-128 bytes. This will not be displayed to the user, use for your internal processes.
    :param provider_token: Payments provider token, obtained via @Botfather
    :param currency: Three-letter ISO 4217 currency code, see https://core.telegram.org/bots/payments#supported-currencies
    :param prices: Price breakdown, a list of components (e.g. product price, tax, discount, delivery cost, delivery tax, bonus, etc.)
    :param start_parameter: Unique deep-linking parameter that can be used to generate this invoice when used as a start parameter
    :param photo_url: URL of the product photo for the invoice. Can be a photo of the goods or a marketing image for a service. People like it better when they see what they are paying for.
    :param photo_size: Photo size
    :param photo_width: Photo width
    :param photo_height: Photo height
    :param need_name: Pass True, if you require the user's full name to complete the order
    :param need_phone_number: Pass True, if you require the user's phone number to complete the order
    :param need_email: Pass True, if you require the user's email to complete the order
    :param need_shipping_address: Pass True, if you require the user's shipping address to complete the order
    :param is_flexible: Pass True, if the final price depends on the shipping method
    :param send_phone_number_to_provider: Pass True, if user's phone number should be sent to provider
    :param send_email_to_provider: Pass True, if user's email address should be sent to provider
    :param disable_notification: Sends the message silently. Users will receive a notification with no sound.
    :param reply_markup: A JSON-serialized object for an inline keyboard. If empty, one 'Pay total price' button will be shown. If not empty, the first button must be a Pay button
    :param provider_data: A JSON-serialized data about the invoice, which will be shared with the payment provider. A detailed description of required fields should be provided by the payment provider.
    :param timeout:
    :param max_tip_amount: The maximum accepted amount for tips in the smallest units of the currency
    :param suggested_tip_amounts: A JSON-serialized array of suggested amounts of tips in the smallest units of the currency.
        At most 4 suggested tip amounts can be specified. The suggested tip amounts must be positive, passed in a strictly increased order and must not exceed max_tip_amount.
    :param protect_content: Protects the contents of the sent message from forwarding and saving
    :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for forum supergroups only
    :param reply_parameters: A JSON-serialized object for an inline keyboard. If empty, one 'Pay total price' button will be shown. If not empty, the first button must be a Pay button.
    :return:
    """
    method_url = r'sendInvoice'
    payload = {'chat_id': chat_id, 'title': title, 'description': description, 'payload': invoice_payload,
                'currency': currency,
                'prices': await _convert_list_json_serializable(prices)}
    if start_parameter:
        payload["start_parameter"] = start_parameter
    if photo_url:
        payload["photo_url"] = photo_url
    if photo_size:
        payload["photo_size"] = photo_size
    if photo_width:
        payload["photo_width"] = photo_width
    if photo_height:
        payload["photo_height"] = photo_height
    if need_name is not None:
        payload["need_name"] = need_name
    if need_phone_number is not None:
        payload["need_phone_number"] = need_phone_number
    if need_email is not None:
        payload["need_email"] = need_email
    if need_shipping_address is not None:
        payload["need_shipping_address"] = need_shipping_address
    if send_phone_number_to_provider is not None:
        payload["send_phone_number_to_provider"] = send_phone_number_to_provider
    if send_email_to_provider is not None:
        payload["send_email_to_provider"] = send_email_to_provider
    if is_flexible is not None:
        payload["is_flexible"] = is_flexible
    if disable_notification is not None:
        payload["disable_notification"] = disable_notification
    if reply_parameters is not None:
        payload["reply_parameters"] = json.dumps(reply_parameters.to_dict())
    if reply_markup:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    if provider_data:
        payload["provider_data"] = provider_data
    if timeout:
        payload["timeout"] = timeout
    if max_tip_amount is not None:
        payload["max_tip_amount"] = max_tip_amount
    if suggested_tip_amounts is not None:
        payload["suggested_tip_amounts"] = json.dumps(suggested_tip_amounts)
    if protect_content is not None:
        payload["protect_content"] = protect_content
    if message_thread_id:
        payload['message_thread_id'] = message_thread_id
    if message_effect_id:
        payload['message_effect_id'] = message_effect_id
    if provider_token is not None:
        payload['provider_token'] = provider_token
    return await _process_request(token, method_url, params=payload)


async def answer_shipping_query(
    token, shipping_query_id, ok, shipping_options=None, error_message=None
):
    """
    If you sent an invoice requesting a shipping address and the parameter is_flexible was specified, the Bot API will send an Update with a shipping_query field to the bot. Use this method to reply to shipping queries. On success, True is returned.
    :param token: Bot's token (you don't need to fill this)
    :param shipping_query_id: Unique identifier for the query to be answered
    :param ok: Specify True if delivery to the specified address is possible and False if there are any problems (for example, if delivery to the specified address is not possible)
    :param shipping_options: Required if ok is True. A JSON-serialized array of available shipping options.
    :param error_message: Required if ok is False. Error message in human readable form that explains why it is impossible to complete the order (e.g. "Sorry, delivery to your desired address is unavailable'). Telegram will display this message to the user.
    :return:
    """
    method_url = "answerShippingQuery"
    payload = {"shipping_query_id": shipping_query_id, "ok": ok}
    if shipping_options:
        payload["shipping_options"] = await _convert_list_json_serializable(
            shipping_options
        )
    if error_message:
        payload["error_message"] = error_message
    return await _process_request(token, method_url, params=payload)


async def answer_pre_checkout_query(
    token, pre_checkout_query_id, ok, error_message=None
):
    """
    Once the user has confirmed their payment and shipping details, the Bot API sends the final confirmation in the form of an Update with the field pre_checkout_query. Use this method to respond to such pre-checkout queries. On success, True is returned. Note: The Bot API must receive an answer within 10 seconds after the pre-checkout query was sent.
    :param token: Bot's token (you don't need to fill this)
    :param pre_checkout_query_id: Unique identifier for the query to be answered
    :param ok: Specify True if everything is alright (goods are available, etc.) and the bot is ready to proceed with the order. Use False if there are any problems.
    :param error_message: Required if ok is False. Error message in human readable form that explains the reason for failure to proceed with the checkout (e.g. "Sorry, somebody just bought the last of our amazing black T-shirts while you were busy filling out your payment details. Please choose a different color or garment!"). Telegram will display this message to the user.
    :return:
    """
    method_url = "answerPreCheckoutQuery"
    payload = {"pre_checkout_query_id": pre_checkout_query_id, "ok": ok}
    if error_message:
        payload["error_message"] = error_message
    return await _process_request(token, method_url, params=payload)


async def refund_star_payment(token, user_id, telegram_payment_charge_id):
    method_url = 'refundStarPayment'
    payload = {'user_id': user_id, 'telegram_payment_charge_id': telegram_payment_charge_id}
    return await _process_request(token, method_url, params=payload)


async def unpin_all_general_forum_topic_messages(token, chat_id):
    method_url = 'unpinAllGeneralForumTopicMessages'
    payload = {'chat_id': chat_id}
    return await _process_request(token, method_url, params=payload)

# InlineQuery


async def answer_callback_query(
    token, callback_query_id, text=None, show_alert=None, url=None, cache_time=None
):
    """
    Use this method to send answers to callback queries sent from inline keyboards. The answer will be displayed to the user as a notification at the top of the chat screen or as an alert. On success, True is returned.
    Alternatively, the user can be redirected to the specified Game URL. For this option to work, you must first create a game for your bot via BotFather and accept the terms. Otherwise, you may use links like telegram.me/your_bot?start=XXXX that open your bot with a parameter.
    :param token: Bot's token (you don't need to fill this)
    :param callback_query_id: Unique identifier for the query to be answered
    :param text: (Optional) Text of the notification. If not specified, nothing will be shown to the user, 0-200 characters
    :param show_alert: (Optional) If true, an alert will be shown by the client instead of a notification at the top of the chat screen. Defaults to false.
    :param url: (Optional) URL that will be opened by the user's client. If you have created a Game and accepted the conditions via @Botfather, specify the URL that opens your game – note that this will only work if the query comes from a callback_game button.
    Otherwise, you may use links like telegram.me/your_bot?start=XXXX that open your bot with a parameter.
    :param cache_time: (Optional) The maximum amount of time in seconds that the result of the callback query may be cached client-side. Telegram apps will support caching starting in version 3.14. Defaults to 0.
    :return:
    """
    method_url = "answerCallbackQuery"
    payload = {"callback_query_id": callback_query_id}
    if text:
        payload["text"] = text
    if show_alert is not None:
        payload["show_alert"] = show_alert
    if url:
        payload["url"] = url
    if cache_time is not None:
        payload["cache_time"] = cache_time
    return await _process_request(token, method_url, params=payload, method="post")


async def get_user_chat_boosts(token, chat_id, user_id):
    method_url = "getUserChatBoosts"
    payload = {"chat_id": chat_id, "user_id": user_id}
    return await _process_request(token, method_url, params=payload)


async def answer_inline_query(
    token,
    inline_query_id,
    results,
    cache_time=None,
    is_personal=None,
    next_offset=None,
    button=None,
):
    method_url = "answerInlineQuery"
    payload = {
        "inline_query_id": inline_query_id,
        "results": await _convert_list_json_serializable(results),
    }
    if cache_time is not None:
        payload["cache_time"] = cache_time
    if is_personal is not None:
        payload["is_personal"] = is_personal
    if next_offset is not None:
        payload["next_offset"] = next_offset
    if button is not None:
        payload["button"] = button.to_json()

    return await _process_request(token, method_url, params=payload, method="post")


async def get_sticker_set(token, name):
    method_url = "getStickerSet"
    return await _process_request(token, method_url, params={"name": name})


async def get_custom_emoji_stickers(token, custom_emoji_ids):
    method_url = r"getCustomEmojiStickers"
    return await _process_request(
        token, method_url, params={"custom_emoji_ids": json.dumps(custom_emoji_ids)}
    )


async def set_sticker_keywords(token, sticker, keywords=None):
    method_url = "setStickerKeywords"
    payload = {"sticker": sticker}
    if keywords:
        payload["keywords"] = json.dumps(keywords)

    return await _process_request(token, method_url, params=payload, method="post")


async def set_sticker_mask_position(token, sticker, mask_position=None):
    method_url = "setStickerMaskPosition"
    payload = {"sticker": sticker}
    if mask_position:
        payload["mask_position"] = mask_position.to_json()
    return await _process_request(token, method_url, params=payload, method="post")


async def upload_sticker_file(token, user_id, sticker, sticker_format):
    method_url = "uploadStickerFile"
    payload = {"user_id": user_id, "sticker_format": sticker_format}
    files = {"sticker": sticker}
    return await _process_request(
        token, method_url, params=payload, files=files, method="post"
    )


async def set_sticker_emoji_list(token, sticker, emoji_list):
    method_url = "setStickerEmojiList"
    payload = {"sticker": sticker, "emoji_list": json.dumps(emoji_list)}
    return await _process_request(token, method_url, params=payload, method="post")


async def delete_sticker_set(token, name):
    method_url = "deleteStickerSet"
    payload = {"name": name}
    return await _process_request(token, method_url, params=payload, method="post")


async def set_custom_emoji_sticker_set_thumbnail(token, name, custom_emoji_id=None):
    method_url = "setCustomEmojiStickerSetThumbnail"
    payload = {"name": name}
    if custom_emoji_id is not None:
        payload["custom_emoji_id"] = custom_emoji_id
    return await _process_request(token, method_url, params=payload, method="post")


async def set_sticker_set_title(token, name, title):
    method_url = "setStickerSetTitle"
    payload = {"name": name, "title": title}
    return await _process_request(token, method_url, params=payload, method="post")


async def create_new_sticker_set(
    token, user_id, name, title, stickers, sticker_type=None, needs_repainting=None
):
    method_url = "createNewStickerSet"
    payload = {"user_id": user_id, "name": name, "title": title}
    if sticker_type:
        payload["sticker_type"] = sticker_type
    if needs_repainting is not None:
        payload["needs_repainting"] = needs_repainting

    files = {}
    lst = []

    for sticker in stickers:
        json_dict, file = sticker.convert_input_sticker()
        json_dict = sticker.to_dict()

        if file:
            list_keys = list(file.keys())
            files[list_keys[0]] = file[list_keys[0]]
        lst.append(json_dict)

    payload["stickers"] = json.dumps(lst)

    return await _process_request(
        token, method_url, params=payload, files=files, method="post"
    )


async def add_sticker_to_set(token, user_id, name, sticker):
    method_url = "addStickerToSet"
    json_dict, files = sticker.convert_input_sticker()
    payload = {"user_id": user_id, "name": name, "sticker": json_dict}

    return await _process_request(
        token, method_url, params=payload, files=files, method="post"
    )


async def set_sticker_position_in_set(token, sticker, position):
    method_url = "setStickerPositionInSet"
    payload = {"sticker": sticker, "position": position}
    return await _process_request(token, method_url, params=payload, method="post")


async def delete_sticker_from_set(token, sticker):
    method_url = 'deleteStickerFromSet'
    payload = {'sticker': sticker}
    return await _process_request(token, method_url, params=payload, method='post')



async def create_invoice_link(token, title, description, payload, provider_token,
            currency, prices, max_tip_amount=None, suggested_tip_amounts=None, provider_data=None,
            photo_url=None, photo_size=None, photo_width=None, photo_height=None, need_name=None, need_phone_number=None,
            need_email=None, need_shipping_address=None, send_phone_number_to_provider=None,
            send_email_to_provider=None, is_flexible=None):
    method_url = r'createInvoiceLink'
    payload = {'title': title, 'description': description, 'payload': payload,
                'currency': currency, 'prices': await _convert_list_json_serializable(prices)}
    if max_tip_amount:
        payload["max_tip_amount"] = max_tip_amount
    if suggested_tip_amounts:
        payload["suggested_tip_amounts"] = json.dumps(suggested_tip_amounts)
    if provider_data:
        payload["provider_data"] = provider_data
    if photo_url:
        payload["photo_url"] = photo_url
    if photo_size:
        payload["photo_size"] = photo_size
    if photo_width:
        payload["photo_width"] = photo_width
    if photo_height:
        payload["photo_height"] = photo_height
    if need_name is not None:
        payload["need_name"] = need_name
    if need_phone_number is not None:
        payload["need_phone_number"] = need_phone_number
    if need_email is not None:
        payload["need_email"] = need_email
    if need_shipping_address is not None:
        payload["need_shipping_address"] = need_shipping_address
    if send_phone_number_to_provider is not None:
        payload["send_phone_number_to_provider"] = send_phone_number_to_provider
    if send_email_to_provider is not None:
        payload["send_email_to_provider"] = send_email_to_provider
    if is_flexible is not None:
        payload['is_flexible'] = is_flexible
    if provider_token is not None:
        payload['provider_token'] = provider_token
    return await _process_request(token, method_url, params=payload, method='post')



# noinspection PyShadowingBuiltins
async def send_poll(
        token, chat_id, question, options,
        is_anonymous = None, type = None, allows_multiple_answers = None, correct_option_id = None,
        explanation = None, explanation_parse_mode=None, open_period = None, close_date = None, is_closed = None,
        disable_notification=False,  
        reply_markup=None, timeout=None, explanation_entities=None, protect_content=None, message_thread_id=None,
        reply_parameters=None,business_connection_id=None, question_parse_mode=None, question_entities=None, message_effect_id=None):
    method_url = r'sendPoll'
    payload = {
        'chat_id': str(chat_id),
        'question': question,
        'options': json.dumps([option.to_dict() for option in options])
    }

    if is_anonymous is not None:
        payload["is_anonymous"] = is_anonymous
    if type is not None:
        payload["type"] = type
    if allows_multiple_answers is not None:
        payload["allows_multiple_answers"] = allows_multiple_answers
    if correct_option_id is not None:
        payload["correct_option_id"] = correct_option_id
    if explanation is not None:
        payload["explanation"] = explanation
    if explanation_parse_mode is not None:
        payload["explanation_parse_mode"] = explanation_parse_mode
    if open_period is not None:
        payload["open_period"] = open_period
    if close_date is not None:
        if isinstance(close_date, datetime):
            payload["close_date"] = close_date.timestamp()
        else:
            payload["close_date"] = close_date
    if is_closed is not None:
        payload['is_closed'] = is_closed
    if disable_notification:
        payload["disable_notification"] = disable_notification
    if reply_parameters is not None:
        payload["reply_parameters"] = json.dumps(reply_parameters.to_dict())
    if reply_markup is not None:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    if timeout:
        payload["timeout"] = timeout
    if explanation_entities:
        payload["explanation_entities"] = json.dumps(
            types.MessageEntity.to_list_of_dicts(explanation_entities)
        )
    if protect_content:
        payload["protect_content"] = protect_content
    if message_thread_id:
        payload["message_thread_id"] = message_thread_id
    if business_connection_id:
        payload['business_connection_id'] = business_connection_id
    if question_parse_mode:
        payload['question_parse_mode'] = question_parse_mode
    if question_entities:
        payload['question_entities'] = json.dumps(types.MessageEntity.to_list_of_dicts(question_entities))
    if message_effect_id:
        payload['message_effect_id'] = message_effect_id
    return await _process_request(token, method_url, params=payload)


async def create_forum_topic(
    token, chat_id, name, icon_color=None, icon_custom_emoji_id=None
):
    method_url = r"createForumTopic"
    payload = {"chat_id": chat_id, "name": name}
    if icon_color:
        payload["icon_color"] = icon_color
    if icon_custom_emoji_id:
        payload["icon_custom_emoji_id"] = icon_custom_emoji_id
    return await _process_request(token, method_url, params=payload)


async def edit_forum_topic(
    token, chat_id, message_thread_id, name=None, icon_custom_emoji_id=None
):
    method_url = r"editForumTopic"
    payload = {"chat_id": chat_id, "message_thread_id": message_thread_id}
    if name is not None:
        payload["name"] = name
    if icon_custom_emoji_id is not None:
        payload["icon_custom_emoji_id"] = icon_custom_emoji_id
    return await _process_request(token, method_url, params=payload)


async def close_forum_topic(token, chat_id, message_thread_id):
    method_url = r"closeForumTopic"
    payload = {"chat_id": chat_id, "message_thread_id": message_thread_id}
    return await _process_request(token, method_url, params=payload)


async def reopen_forum_topic(token, chat_id, message_thread_id):
    method_url = r"reopenForumTopic"
    payload = {"chat_id": chat_id, "message_thread_id": message_thread_id}
    return await _process_request(token, method_url, params=payload)


async def delete_forum_topic(token, chat_id, message_thread_id):
    method_url = r"deleteForumTopic"
    payload = {"chat_id": chat_id, "message_thread_id": message_thread_id}
    return await _process_request(token, method_url, params=payload)


async def unpin_all_forum_topic_messages(token, chat_id, message_thread_id):
    method_url = r"unpinAllForumTopicMessages"
    payload = {"chat_id": chat_id, "message_thread_id": message_thread_id}
    return await _process_request(token, method_url, params=payload)


async def get_forum_topic_icon_stickers(token):
    method_url = r"getForumTopicIconStickers"
    return await _process_request(token, method_url)


async def edit_general_forum_topic(token, chat_id, name):
    method_url = r"editGeneralForumTopic"
    payload = {"chat_id": chat_id, "name": name}
    return await _process_request(token, method_url, params=payload)


async def close_general_forum_topic(token, chat_id):
    method_url = r"closeGeneralForumTopic"
    payload = {"chat_id": chat_id}
    return await _process_request(token, method_url, params=payload)


async def reopen_general_forum_topic(token, chat_id):
    method_url = r"reopenGeneralForumTopic"
    payload = {"chat_id": chat_id}
    return await _process_request(token, method_url, params=payload)


async def hide_general_forum_topic(token, chat_id):
    method_url = r"hideGeneralForumTopic"
    payload = {"chat_id": chat_id}
    return await _process_request(token, method_url, params=payload)


async def unhide_general_forum_topic(token, chat_id):
    method_url = r"unhideGeneralForumTopic"
    payload = {"chat_id": chat_id}
    return await _process_request(token, method_url, params=payload)


async def delete_messages(token, chat_id, message_ids):
    method_url = "deleteMessages"
    payload = {"chat_id": chat_id, "message_ids": json.dumps(message_ids)}
    return await _process_request(token, method_url, params=payload)


async def forward_messages(
    token,
    chat_id,
    from_chat_id,
    message_ids,
    disable_notification=None,
    message_thread_id=None,
    protect_content=None,
):
    method_url = "forwardMessages"
    payload = {
        "chat_id": chat_id,
        "from_chat_id": from_chat_id,
        "message_ids": json.dumps(message_ids),
    }
    if disable_notification is not None:
        payload["disable_notification"] = disable_notification
    if message_thread_id is not None:
        payload["message_thread_id"] = message_thread_id
    if protect_content is not None:
        payload["protect_content"] = protect_content

    result = await _process_request(token, method_url, params=payload)
    return result


async def copy_messages(
    token,
    chat_id,
    from_chat_id,
    message_ids,
    disable_notification=None,
    message_thread_id=None,
    protect_content=None,
    remove_caption=None,
):
    method_url = "copyMessages"
    payload = {
        "chat_id": chat_id,
        "from_chat_id": from_chat_id,
        "message_ids": json.dumps(message_ids),
    }
    if disable_notification is not None:
        payload["disable_notification"] = disable_notification
    if message_thread_id is not None:
        payload["message_thread_id"] = message_thread_id
    if protect_content is not None:
        payload["protect_content"] = protect_content
    if remove_caption is not None:
        payload["remove_caption"] = remove_caption

    result = await _process_request(token, method_url, params=payload)
    return result


async def _convert_list_json_serializable(results):
    ret = ""
    for r in results:
        if isinstance(r, types.JsonSerializable):
            ret = ret + r.to_json() + ","
    if len(ret) > 0:
        ret = ret[:-1]
    return "[" + ret + "]"


async def convert_input_media(media):
    if isinstance(media, types.InputMedia):
        return media.convert_input_media()
    return None, None


async def convert_input_media_array(array):
    media = []
    files = {}
    for input_media in array:
        if isinstance(input_media, types.InputMedia):
            media_dict = input_media.to_dict()
            if media_dict["media"].startswith("attach://"):
                key = media_dict["media"].replace("attach://", "")
                files[key] = input_media.media
            media.append(media_dict)
    return json.dumps(media), files


async def _no_encode(func):
    def wrapper(key, val):
        if key == "filename":
            return "{0}={1}".format(key, val)
        else:
            return func(key, val)

    return wrapper


async def stop_poll(token, chat_id, message_id, reply_markup=None):
    method_url = r"stopPoll"
    payload = {"chat_id": str(chat_id), "message_id": message_id}
    if reply_markup:
        payload["reply_markup"] = await _convert_markup(reply_markup)
    return await _process_request(token, method_url, params=payload)


# exceptions
class ApiException(Exception):
    """
    This class represents a base Exception thrown when a call to the Telegram API fails.
    In addition to an informative message, it has a `function_name` and a `result` attribute, which respectively
    contain the name of the failed function and the returned result that made the function to be considered  as
    failed.
    """

    def __init__(self, msg, function_name, result):
        super(ApiException, self).__init__(
            "A request to the Telegram API was unsuccessful. {0}".format(msg)
        )
        self.function_name = function_name
        self.result = result


class ApiHTTPException(ApiException):
    """
    This class represents an Exception thrown when a call to the
    Telegram API server returns HTTP code that is not 200.
    """

    def __init__(self, function_name, result: aiohttp.ClientResponse):
        super(ApiHTTPException, self).__init__(
            "The server returned HTTP {0} {1}. Response body:\n[{2}]".format(
                result.status, result.reason, result.request_info
            ),
            function_name,
            result,
        )


class ApiInvalidJSONException(ApiException):
    """
    This class represents an Exception thrown when a call to the
    Telegram API server returns invalid json.
    """

    def __init__(self, function_name, result):
        super(ApiInvalidJSONException, self).__init__(
            "The server returned an invalid JSON response. Response body:\n[{0}]".format(
                result
            ),
            function_name,
            result,
        )


class ApiTelegramException(ApiException):
    """
    This class represents an Exception thrown when a Telegram API returns error code.
    """

    def __init__(self, function_name, result, result_json):
        super(ApiTelegramException, self).__init__(
            "Error code: {0}. Description: {1}".format(
                result_json["error_code"], result_json["description"]
            ),
            function_name,
            result,
        )
        self.result_json = result_json
        self.error_code = result_json["error_code"]
        self.description = result_json["description"]


class RequestTimeout(Exception):
    """
    This class represents a request timeout.
    """

    pass



# -*- coding: utf-8 -*-
"""
This Example will show you usage of TextFilter
In this example you will see how to use TextFilter
with (message_handler, callback_query_handler, poll_handler)
"""
import asyncio

from telebot import types
from telebot.async_telebot import AsyncTeleBot
from telebot.asyncio_filters import IsReplyFilter, TextFilter, TextMatchFilter

bot = AsyncTeleBot("")


@bot.message_handler(text=TextFilter(equals="hello"))
async def hello_handler(message: types.Message):
    await bot.send_message(message.chat.id, message.text)


@bot.message_handler(text=TextFilter(equals="hello", ignore_case=True))
async def hello_handler_ignore_case(message: types.Message):
    await bot.send_message(message.chat.id, message.text + " ignore case")


@bot.message_handler(text=TextFilter(contains=["good", "bad"]))
async def contains_handler(message: types.Message):
    await bot.send_message(message.chat.id, message.text)


@bot.message_handler(text=TextFilter(contains=["good", "bad"], ignore_case=True))
async def contains_handler_ignore_case(message: types.Message):
    await bot.send_message(message.chat.id, message.text + " ignore case")


@bot.message_handler(text=TextFilter(starts_with="st"))  # stArk, steve, stONE
async def starts_with_handler(message: types.Message):
    await bot.send_message(message.chat.id, message.text)


@bot.message_handler(
    text=TextFilter(starts_with="st", ignore_case=True)
)  # STark, sTeve, stONE
async def starts_with_handler_ignore_case(message: types.Message):
    await bot.send_message(message.chat.id, message.text + " ignore case")


@bot.message_handler(text=TextFilter(ends_with="ay"))  # wednesday, SUNday, WeekDay
async def ends_with_handler(message: types.Message):
    await bot.send_message(message.chat.id, message.text)


@bot.message_handler(
    text=TextFilter(ends_with="ay", ignore_case=True)
)  # wednesdAY, sundAy, WeekdaY
async def ends_with_handler_ignore_case(message: types.Message):
    await bot.send_message(message.chat.id, message.text + " ignore case")


@bot.message_handler(text=TextFilter(equals="/callback"))
async def send_callback(message: types.Message):
    keyboard = types.InlineKeyboardMarkup(
        keyboard=[
            [types.InlineKeyboardButton(text="callback data", callback_data="example")],
            [
                types.InlineKeyboardButton(
                    text="ignore case callback data", callback_data="ExAmPLe"
                )
            ],
        ]
    )
    await bot.send_message(message.chat.id, message.text, reply_markup=keyboard)


@bot.callback_query_handler(func=None, text=TextFilter(equals="example"))
async def callback_query_handler(call: types.CallbackQuery):
    await bot.answer_callback_query(call.id, call.data, show_alert=True)


@bot.callback_query_handler(
    func=None, text=TextFilter(equals="example", ignore_case=True)
)
async def callback_query_handler_ignore_case(call: types.CallbackQuery):
    await bot.answer_callback_query(
        call.id, call.data + " ignore case", show_alert=True
    )


@bot.message_handler(text=TextFilter(equals="/poll"))
async def send_poll(message: types.Message):
    await bot.send_poll(
        message.chat.id,
        question="When do you prefer to work?",
        options=["Morning", "Night"],
    )
    await bot.send_poll(
        message.chat.id,
        question="WHEN DO you pRefeR to worK?",
        options=["Morning", "Night"],
    )


@bot.poll_handler(func=None, text=TextFilter(equals="When do you prefer to work?"))
async def poll_question_handler(poll: types.Poll):
    print(poll.question)


@bot.poll_handler(
    func=None, text=TextFilter(equals="When do you prefer to work?", ignore_case=True)
)
async def poll_question_handler_ignore_case(poll: types.Poll):
    print(poll.question + " ignore case")


# either hi or contains one of (привет, salom)
@bot.message_handler(
    text=TextFilter(equals="hi", contains=("привет", "salom"), ignore_case=True)
)
async def multiple_patterns_handler(message: types.Message):
    await bot.send_message(message.chat.id, message.text)


# starts with one of (mi, mea) for ex. minor, milk, meal, meat
@bot.message_handler(text=TextFilter(starts_with=["mi", "mea"], ignore_case=True))
async def multiple_starts_with_handler(message: types.Message):
    await bot.send_message(message.chat.id, message.text)


# ends with one of (es, on) for ex. Jones, Davies, Johnson, Wilson
@bot.message_handler(text=TextFilter(ends_with=["es", "on"], ignore_case=True))
async def multiple_ends_with_handler(message: types.Message):
    await bot.send_message(message.chat.id, message.text)


# !ban /ban .ban !бан /бан .бан
@bot.message_handler(
    is_reply=True,
    text=TextFilter(
        starts_with=("!", "/", "."), ends_with=["ban", "бан"], ignore_case=True
    ),
)
async def ban_command_handler(message: types.Message):
    if len(message.text) == 4 and message.chat.type != "private":
        try:
            await bot.ban_chat_member(
                message.chat.id, message.reply_to_message.from_user.id
            )
            await bot.reply_to(message.reply_to_message, "Banned.")
        except Exception as err:
            print(err.args)
            return


if __name__ == "__main__":
    bot.add_custom_filter(TextMatchFilter())
    bot.add_custom_filter(IsReplyFilter())
    asyncio.run(bot.polling())



import telebot
from telebot.async_telebot import AsyncTeleBot

bot = AsyncTeleBot("TOKEN")


# Check if message starts with @admin tag
@bot.message_handler(text_startswith="@admin")
async def start_filter(message):
    await bot.send_message(message.chat.id, "Looks like you are calling admin, wait...")


# Check if text is hi or hello
@bot.message_handler(text=["hi", "hello"])
async def text_filter(message):
    await bot.send_message(
        message.chat.id, "Hi, {name}!".format(name=message.from_user.first_name)
    )


# Do not forget to register filters
bot.add_custom_filter(telebot.asyncio_filters.TextMatchFilter())
bot.add_custom_filter(telebot.asyncio_filters.TextStartsFilter())

import asyncio

asyncio.run(bot.polling())



from telebot import asyncio_filters
from telebot.async_telebot import AsyncTeleBot

# new feature for states.
from telebot.asyncio_handler_backends import State, StatesGroup

# list of storages, you can use any storage
from telebot.asyncio_storage import StateMemoryStorage

# default state storage is statememorystorage
bot = AsyncTeleBot("TOKEN", state_storage=StateMemoryStorage())


# Just create different statesgroup
class MyStates(StatesGroup):
    name = State()  # statesgroup should contain states
    surname = State()
    age = State()


# set_state -> sets a new state
# delete_state -> delets state if exists
# get_state -> returns state if exists


@bot.message_handler(commands=["start"])
async def start_ex(message):
    """
    Start command. Here we are starting state
    """
    await bot.set_state(message.from_user.id, MyStates.name, message.chat.id)
    await bot.send_message(message.chat.id, "Hi, write me a name")


@bot.message_handler(state="*", commands="cancel")
async def any_state(message):
    """
    Cancel state
    """
    await bot.send_message(message.chat.id, "Your state was cancelled.")
    await bot.delete_state(message.from_user.id, message.chat.id)


@bot.message_handler(state=MyStates.name)
async def name_get(message):
    """
    State 1. Will process when user's state is MyStates.name.
    """
    await bot.send_message(message.chat.id, f"Now write me a surname")
    await bot.set_state(message.from_user.id, MyStates.surname, message.chat.id)
    async with bot.retrieve_data(message.from_user.id, message.chat.id) as data:
        data["name"] = message.text


@bot.message_handler(state=MyStates.surname)
async def ask_age(message):
    """
    State 2. Will process when user's state is MyStates.surname.
    """
    await bot.send_message(message.chat.id, "What is your age?")
    await bot.set_state(message.from_user.id, MyStates.age, message.chat.id)
    async with bot.retrieve_data(message.from_user.id, message.chat.id) as data:
        data["surname"] = message.text


# result
@bot.message_handler(state=MyStates.age, is_digit=True)
async def ready_for_answer(message):
    """
    State 3. Will process when user's state is MyStates.age.
    """
    async with bot.retrieve_data(message.from_user.id, message.chat.id) as data:
        await bot.send_message(
            message.chat.id,
            "Ready, take a look:\n<b>Name: {name}\nSurname: {surname}\nAge: {age}</b>".format(
                name=data["name"], surname=data["surname"], age=message.text
            ),
            parse_mode="html",
        )
    await bot.delete_state(message.from_user.id, message.chat.id)


# incorrect number
@bot.message_handler(state=MyStates.age, is_digit=False)
async def age_incorrect(message):
    """
    Will process for wrong input when state is MyState.age
    """
    await bot.send_message(
        message.chat.id,
        "Looks like you are submitting a string in the field age. Please enter a number",
    )


# register filters

bot.add_custom_filter(asyncio_filters.StateFilter(bot))
bot.add_custom_filter(asyncio_filters.IsDigitFilter())


import asyncio

asyncio.run(bot.polling())



import copy


class StateStorageBase:
    def __init__(self) -> None:
        pass

    def set_data(self, chat_id, user_id, key, value):
        """
        Set data for a user in a particular chat.
        """
        raise NotImplementedError

    def get_data(self, chat_id, user_id):
        """
        Get data for a user in a particular chat.
        """
        raise NotImplementedError

    def set_state(self, chat_id, user_id, state):
        """
        Set state for a particular user.

        ! Note that you should create a
        record if it does not exist, and
        if a record with state already exists,
        you need to update a record.
        """
        raise NotImplementedError

    def delete_state(self, chat_id, user_id):
        """
        Delete state for a particular user.
        """
        raise NotImplementedError

    def reset_data(self, chat_id, user_id):
        """
        Reset data for a particular user in a chat.
        """
        raise NotImplementedError

    def get_state(self, chat_id, user_id):
        raise NotImplementedError

    def get_interactive_data(self, chat_id, user_id):
        raise NotImplementedError

    def save(self, chat_id, user_id, data):
        raise NotImplementedError


class StateContext:
    """
    Class for data.
    """

    def __init__(self, obj, chat_id, user_id) -> None:
        self.obj = obj
        self.data = copy.deepcopy(obj.get_data(chat_id, user_id))
        self.chat_id = chat_id
        self.user_id = user_id

    def __enter__(self):
        return self.data

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.obj.save(self.chat_id, self.user_id, self.data)



import telebot
from telebot import custom_filters

bot = telebot.TeleBot("token")


# Chat id can be private or supergroups.
@bot.message_handler(
    chat_id=[12345678], commands=["admin"]
)  # chat_id checks id corresponds to your list or not.
def admin_rep(message):
    bot.send_message(message.chat.id, "You are allowed to use this command.")


@bot.message_handler(commands=["admin"])
def not_admin(message):
    bot.send_message(message.chat.id, "You are not allowed to use this command")


# Do not forget to register
bot.add_custom_filter(custom_filters.ChatFilter())

bot.infinity_polling()



import telebot
from telebot import types, util

bot = telebot.TeleBot("token")


# chat_member_handler. When status changes, telegram gives update. check status from old_chat_member and new_chat_member.
@bot.chat_member_handler()
def chat_m(message: types.ChatMemberUpdated):
    old = message.old_chat_member
    new = message.new_chat_member
    if new.status == "member":
        bot.send_message(
            message.chat.id, "Hello {name}!".format(name=new.user.first_name)
        )  # Welcome message


# if bot is added to group, this handler will work
@bot.my_chat_member_handler()
def my_chat_m(message: types.ChatMemberUpdated):
    old = message.old_chat_member
    new = message.new_chat_member
    if new.status == "member":
        bot.send_message(
            message.chat.id, "Somebody added me to group"
        )  # Welcome message, if bot was added to group
        bot.leave_chat(message.chat.id)


# content_Type_service is:
#'new_chat_members', 'left_chat_member', 'new_chat_title', 'new_chat_photo', 'delete_chat_photo', 'group_chat_created',
#'supergroup_chat_created', 'channel_chat_created', 'migrate_to_chat_id', 'migrate_from_chat_id', 'pinned_message',
#'proximity_alert_triggered', 'video_chat_scheduled', 'video_chat_started', 'video_chat_ended',
#'video_chat_participants_invited', 'message_auto_delete_timer_changed'
# this handler deletes service messages


@bot.message_handler(content_types=util.content_type_service)
def delall(message: types.Message):
    bot.delete_message(message.chat.id, message.message_id)


bot.infinity_polling(allowed_updates=util.update_types)



import telebot
from telebot import custom_filters

bot = telebot.TeleBot("TOKEN")


# Handler
@bot.message_handler(chat_types=["supergroup"], is_chat_admin=True)
def answer_for_admin(message):
    bot.send_message(message.chat.id, "hello my admin")


# Register filter
bot.add_custom_filter(custom_filters.IsAdminFilter(bot))
bot.infinity_polling()



"""
Example of running PyTelegramBotAPI serverless in Amazon AWS Lambdaю
You have to set your lambda's url as telegram webhook manually https://core.telegram.org/bots/api#setwebhook
"""

import json
import logging
import os

import telebot

API_TOKEN = os.environ["TELEGRAM_TOKEN"]


logger = telebot.logger
telebot.logger.setLevel(logging.INFO)

bot = telebot.TeleBot(API_TOKEN, threaded=False)


def process_event(event):
    # Get telegram webhook json from event
    request_body_dict = json.loads(event["body"])
    # Parse updates from json
    update = telebot.types.Update.de_json(request_body_dict)
    # Run handlers and etc for updates
    bot.process_new_updates([update])


def lambda_handler(event, context):
    # Process event from aws and respond
    process_event(event)
    return {"statusCode": 200}


# Handle '/start' and '/help'
@bot.message_handler(commands=["help", "start"])
def send_welcome(message):
    bot.reply_to(
        message,
        ("Hi there, I am EchoBot.\n" "I am here to echo your kind words back to you."),
    )


# Handle all other messages
@bot.message_handler(func=lambda message: True, content_types=["text"])
def echo_message(message):
    bot.reply_to(message, message.text)



import telebot  # telebot
from telebot import custom_filters
from telebot.handler_backends import State, StatesGroup  # States

# States storage
from telebot.storage import StateMemoryStorage

# Starting from version 4.4.0+, we support storages.
# StateRedisStorage -> Redis-based storage.
# StatePickleStorage -> Pickle-based storage.
# For redis, you will need to install redis.
# Pass host, db, password, or anything else,
# if you need to change config for redis.
# Pickle requires path. Default path is in folder .state-saves.
# If you were using older version of pytba for pickle,
# you need to migrate from old pickle to new by using
# StatePickleStorage().convert_old_to_new()


# Now, you can pass storage to bot.
state_storage = StateMemoryStorage()  # you can init here another storage

bot = telebot.TeleBot("TOKEN", state_storage=state_storage)


# States group.
class MyStates(StatesGroup):
    # Just name variables differently
    name = State()  # creating instances of State class is enough from now
    surname = State()
    age = State()


@bot.message_handler(commands=["start"])
def start_ex(message):
    """
    Start command. Here we are starting state
    """
    bot.set_state(message.from_user.id, MyStates.name, message.chat.id)
    bot.send_message(message.chat.id, "Hi, write me a name")


# Any state
@bot.message_handler(state="*", commands=["cancel"])
def any_state(message):
    """
    Cancel state
    """
    bot.send_message(message.chat.id, "Your state was cancelled.")
    bot.delete_state(message.from_user.id, message.chat.id)


@bot.message_handler(state=MyStates.name)
def name_get(message):
    """
    State 1. Will process when user's state is MyStates.name.
    """
    bot.send_message(message.chat.id, "Now write me a surname")
    bot.set_state(message.from_user.id, MyStates.surname, message.chat.id)
    with bot.retrieve_data(message.from_user.id, message.chat.id) as data:
        data["name"] = message.text


@bot.message_handler(state=MyStates.surname)
def ask_age(message):
    """
    State 2. Will process when user's state is MyStates.surname.
    """
    bot.send_message(message.chat.id, "What is your age?")
    bot.set_state(message.from_user.id, MyStates.age, message.chat.id)
    with bot.retrieve_data(message.from_user.id, message.chat.id) as data:
        data["surname"] = message.text


# result
@bot.message_handler(state=MyStates.age, is_digit=True)
def ready_for_answer(message):
    """
    State 3. Will process when user's state is MyStates.age.
    """
    with bot.retrieve_data(message.from_user.id, message.chat.id) as data:
        msg = (
            "Ready, take a look:\n<b>"
            f"Name: {data['name']}\n"
            f"Surname: {data['surname']}\n"
            f"Age: {message.text}</b>"
        )
        bot.send_message(message.chat.id, msg, parse_mode="html")
    bot.delete_state(message.from_user.id, message.chat.id)


# incorrect number
@bot.message_handler(state=MyStates.age, is_digit=False)
def age_incorrect(message):
    """
    Wrong response for MyStates.age
    """
    bot.send_message(
        message.chat.id,
        "Looks like you are submitting a string in the field age. Please enter a number",
    )


# register filters

bot.add_custom_filter(custom_filters.StateFilter(bot))
bot.add_custom_filter(custom_filters.IsDigitFilter())

bot.infinity_polling(skip_pending=True)



import telebot

bot = telebot.TeleBot("TOKEN")


# AdvancedCustomFilter is for list, string filter values
class MainFilter(telebot.custom_filters.AdvancedCustomFilter):
    key = "text"

    @staticmethod
    def check(message, text):
        return message.text in text


# SimpleCustomFilter is for boolean values, such as is_admin=True
class IsAdmin(telebot.custom_filters.SimpleCustomFilter):
    key = "is_admin"

    @staticmethod
    def check(message: telebot.types.Message):
        return bot.get_chat_member(message.chat.id, message.from_user.id).status in [
            "administrator",
            "creator",
        ]


@bot.message_handler(is_admin=True, commands=["admin"])  # Check if user is admin
def admin_rep(message):
    bot.send_message(message.chat.id, "Hi admin")


@bot.message_handler(is_admin=False, commands=["admin"])  # If user is not admin
def not_admin(message):
    bot.send_message(message.chat.id, "You are not admin")


@bot.message_handler(text=["hi"])  # Response to hi message
def welcome_hi(message):
    bot.send_message(message.chat.id, "You said hi")


@bot.message_handler(text=["bye"])  # Response to bye message
def bye_user(message):
    bot.send_message(message.chat.id, "You said bye")


# Do not forget to register filters
bot.add_custom_filter(MainFilter())
bot.add_custom_filter(IsAdmin())

bot.infinity_polling(skip_pending=True)  # Skip old updates
